"""Helper functions for testing logging."""

from __future__ import annotations

import json
from collections.abc import Iterable
from datetime import UTC, datetime, timedelta
from logging import getLevelName
from typing import Any

__all__ = ["parse_log_tuples"]


def _assert_timestamp_recent(isotimestamp: str, now: datetime) -> None:
    """Assert if a timestamp is malformed or older than 10 seconds."""
    if not isotimestamp.endswith("Z"):
        raise AssertionError(f'"{isotimestamp}".endswith("Z")')
    timestamp = datetime.fromisoformat(isotimestamp[:-1]).replace(tzinfo=UTC)
    if not (now - timedelta(hours=1) < timestamp < now):
        msg = f"{isotimestamp} within 1 hour of {now!s}"
        raise AssertionError(msg)


def parse_log_tuples(
    logger_name: str,
    record_tuples: Iterable[tuple[str, int, str]],
    *,
    ignore_debug: bool = False,
) -> list[dict[str, Any]]:
    """Parse JSON log record tuples into structured data.

    This function is intended for use in test suites that want to check the
    messages logged via structlog_ in the Safir production profile, meaning
    that each log message is in JSON. The message bodies are parsed, checked
    to ensure they're logged with the provided logger name, stripped of
    uninteresting data after verification (any timestamp, ``request_id``, or
    HTTP user agent, all of which vary in uninteresting ways), and then
    returned as structured records.

    Any timestamp in the log message is expected to be within 10 seconds of
    the current time.

    Parameters
    ----------
    logger_name
        Only look at messages logged by this logger. Any messages from a
        different logger will be ignored. The logger is removed from the
        message before returning it.
    record_tuples
        Tuples of log records consisting of logger name, log level, and
        message. Normally this argument should be ``caplog.record_tuples``,
        provided by the pytest ``caplog`` fixture.
    ignore_debug
        If set to `True`, filter out all debug messages.

    Returns
    -------
    list of dict
        List of parsed JSON dictionaries with the common log attributes
        removed (after validation).

    Raises
    ------
    AssertionError
        Raised if any of the validation of the log messages fails.
    """
    now = datetime.now(tz=UTC)
    messages = []

    for name, level, message_raw in record_tuples:
        message = json.loads(message_raw)

        if name != logger_name:
            continue
        if message["logger"] != logger_name:
            msg = f"logger {message['logger']} == {logger_name}"
            raise AssertionError(msg)
        del message["logger"]

        level_name = getLevelName(level).lower()
        if message["severity"] != level_name:
            msg = f"severity {message['severity']} == {level}"
            raise AssertionError(msg)

        if "timestamp" in message:
            _assert_timestamp_recent(message["timestamp"], now)
            del message["timestamp"]

        if "request_id" in message:
            del message["request_id"]
            if "userAgent" in message["httpRequest"]:
                del message["httpRequest"]["userAgent"]

        if not ignore_debug or message["severity"] != "debug":
            messages.append(message)

    return messages



"""Slack Block Kit message models."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import UTC, datetime
from typing import Any, ClassVar, Self, override

from httpx import HTTPError, HTTPStatusError
from pydantic import BaseModel, field_validator

from safir.datetime import format_datetime_for_logging

from .sentry import SentryEventInfo

__all__ = [
    "SentryEventInfo",
    "SlackBaseBlock",
    "SlackBaseField",
    "SlackCodeBlock",
    "SlackCodeField",
    "SlackException",
    "SlackMessage",
    "SlackTextBlock",
    "SlackTextField",
    "SlackWebException",
]


class SlackBaseBlock(BaseModel, metaclass=ABCMeta):
    """Base class for any Slack Block Kit block."""

    max_formatted_length: ClassVar[int] = 3000
    """Maximum length of formatted output, imposed by Slack.

    Intended to be overridden by child classes that need to impose different
    maximum lengths.
    """

    @abstractmethod
    def to_slack(self) -> dict[str, Any]:
        """Convert to a Slack Block Kit block.

        Returns
        -------
        dict
            A Slack Block Kit block suitable for including in the ``fields``
            or ``text`` section of a ``blocks`` element.
        """


class SlackTextBlock(SlackBaseBlock):
    """A component of a Slack message with a heading and a text body.

    If the formatted output is longer than 3000 characters, it will be
    truncated to avoid the strict uppper limit imposed by Slack.
    """

    heading: str
    """Heading of the field (shown in bold)."""

    text: str
    """Text of the field as normal text.

    This is always marked as vertabim, so channel mentions or @-mentions of
    users will not be treated as special.
    """

    @override
    def to_slack(self) -> dict[str, Any]:
        """Convert to a Slack Block Kit block.

        Returns
        -------
        dict
            A Slack Block Kit block suitable for including in the ``fields``
            or ``text`` section of a ``blocks`` element.
        """
        heading = f"*{self.heading}*\n"
        max_length = self.max_formatted_length - len(heading)
        body = _format_and_truncate_at_end(self.text, max_length)
        return {"type": "mrkdwn", "text": heading + body, "verbatim": True}


class SlackCodeBlock(SlackBaseBlock):
    """A component of a Slack message with a heading and a code block.

    If the formatted output is longer than 3000 characters, it will be
    truncated to avoid the strict upper limit imposed by Slack.
    """

    heading: str
    """Heading of the field (shown in bold)."""

    code: str
    """Text of the field as a code block."""

    @override
    def to_slack(self) -> dict[str, Any]:
        """Convert to a Slack Block Kit block.

        Returns
        -------
        dict
            A Slack Block Kit block suitable for including in the ``fields``
            or ``text`` section of a ``blocks`` element.
        """
        heading = f"*{self.heading}*\n"
        extra_needed = len(heading) + 8  # ```\n\n```
        max_length = self.max_formatted_length - extra_needed
        code = _format_and_truncate_at_start(self.code, max_length)
        text = f"{heading}```\n{code}\n```"
        return {"type": "mrkdwn", "text": text, "verbatim": True}


class SlackBaseField(SlackBaseBlock):
    """Base class for Slack Block Kit blocks for the ``fields`` section."""

    max_formatted_length: ClassVar[int] = 2000


class SlackTextField(SlackTextBlock, SlackBaseField):
    """One field in a Slack message with a heading and text body.

    Intended for use in the ``fields`` portion of a Block Kit message. If the
    formatted output is longer than 2000 characters, it will be truncated to
    avoid the strict upper limit imposed by Slack.
    """


class SlackCodeField(SlackCodeBlock, SlackBaseField):
    """An attachment in a Slack message with a heading and text body.

    Intended for use in the ``fields`` portion of a Block Kit message. If
    the formatted output is longer than 2000 characters, it will be truncated
    to avoid the strict upper limit imposed by Slack.
    """


class SlackMessage(BaseModel):
    """Message to post to Slack.

    The ``message`` attribute will be the initial part of the message.

    All fields in ``fields`` will be shown below that message, formatted in
    two columns. Order of ``fields`` is preserved; they will be laid out left
    to right and then top to bottom in the order given. Then, ``blocks`` will
    be added, if any, in one column below the fields. Finally, ``attachments``
    will be added to the end as attachments, which get somewhat different
    formatting (for example, long attachments are collapsed by default).

    At most ten elements are allowed in ``fields``. They should be used for
    short information, generally a single half-line at most.  Longer
    information should go into ``blocks`` or ``attachments``.
    """

    message: str
    """Main part of the message."""

    verbatim: bool = True
    """Whether the main part of the message should be marked verbatim.

    Verbatim messages in Slack don't expand channel references or create user
    notifications. This is the default, but can be set to `False` to allow
    any such elements in the message to be recognized by Slack. Do not set
    this to `False` with untrusted input.
    """

    fields: list[SlackBaseField] = []
    """Short key/value fields to include in the message (at most 10)."""

    blocks: list[SlackBaseBlock] = []
    """Additional text blocks to include in the message (after fields)."""

    attachments: list[SlackBaseBlock] = []
    """Longer sections to include as attachments.

    Notes
    -----
    Slack has marked attachments as legacy and warns that future changes may
    reduce their visibility or utility. Unfortunately, there is no other way
    to attach possibly-long text where Slack will hide long content by default
    but allow the user to expand it. We therefore continue to use attachments
    for long text for want of a better alternative.
    """

    @field_validator("fields")
    @classmethod
    def _validate_fields(cls, v: list[SlackBaseField]) -> list[SlackBaseField]:
        """Check constraints on fields.

        Slack imposes a maximum of 10 items in a ``fields`` array. Also ensure
        that no fields are actually attachments, since in that case they may
        not be truncated to the correct length. (The type system we're using
        doesn't allow Pydantic to check this directly.)
        """
        if len(v) > 10:
            msg = f"Slack does not allow more than 10 fields ({len(v)} seen)"
            raise ValueError(msg)
        return v

    def to_slack(self) -> dict[str, Any]:
        """Convert to a Slack Block Kit message.

        Returns
        -------
        dict
            A Slack Block Kit data structure suitable for serializing to
            JSON and sending to Slack.
        """
        attachments = [
            {"type": "section", "text": a.to_slack()} for a in self.attachments
        ]
        message = _format_and_truncate_at_end(self.message, 3000)
        blocks: list[dict[str, Any]] = [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": message,
                    "verbatim": self.verbatim,
                },
            }
        ]
        fields = [f.to_slack() for f in self.fields]
        if fields:
            blocks.append({"type": "section", "fields": fields})
        blocks.extend(
            [{"type": "section", "text": b.to_slack()} for b in self.blocks]
        )
        result: dict[str, Any] = {"blocks": blocks}
        if attachments:
            result["attachments"] = [{"blocks": attachments}]
        elif fields or self.blocks:
            result["blocks"].append({"type": "divider"})
        return result


class SlackException(Exception):
    """Parent class of exceptions that can be reported to Slack or Sentry.

    Intended to be subclassed. Subclasses may wish to override the
    ``to_slack`` and ``to_sentry`` methods.

    Attributes
    ----------
    message
        Error message represented by this exception.
    user
        Username associated with the exception.
    failed_at
        When the failure occurred. Defaults to the time the exception was
        created.

    Parameters
    ----------
    message
        Exception string value, which is the default Slack message.
    user
        Identity of user triggering the exception, if known.
    failed_at
        When the exception happened. Omit to use the current time.
    """

    def __init__(
        self,
        message: str,
        user: str | None = None,
        *,
        failed_at: datetime | None = None,
    ) -> None:
        # Do not call the parent Exception constructor here, because calling
        # it with a different number of arguments than the constructor
        # argument of derived exceptions breaks pickling. See the end of
        # https://github.com/python/cpython/issues/44791. This requires
        # implementing __str__ rather than relying on the default behavior.
        #
        # Arguably, this is a bug in the __reduce__ method of BaseException
        # and its interaction with constructors, but it appears to be hard to
        # fix. See https://github.com/python/cpython/issues/76877.
        self.message = message
        self.user = user
        self.failed_at = failed_at if failed_at else datetime.now(tz=UTC)

    def __str__(self) -> str:
        return self.message

    def to_slack(self) -> SlackMessage:
        """Format the exception as a Slack message.

        This is the generic version that only reports the text of the
        exception and the base fields. Child exceptions may want to override
        it to add more metadata.

        Returns
        -------
        SlackMessage
            Slack message suitable for posting with
            `safir.slack.webhook.SlackWebhookClient`.
        """
        failed_at = format_datetime_for_logging(self.failed_at)
        fields: list[SlackBaseField] = [
            SlackTextField(heading="Exception type", text=type(self).__name__),
            SlackTextField(heading="Failed at", text=failed_at),
        ]
        if self.user:
            fields.append(SlackTextField(heading="User", text=self.user))
        return SlackMessage(message=str(self), fields=fields)

    def to_sentry(self) -> SentryEventInfo:
        """Return an object with tags and contexts to add to Sentry events.

        This is the generic version that only adds the username as a tag. Child
        exceptions may want to override it to add more metadata.
        """
        return SentryEventInfo(username=self.user)


class SlackWebException(SlackException):
    """Parent class of exceptions arising from HTTPX_ failures.

    Captures additional information from any HTTPX_ exception.  Intended to be
    subclassed.  Subclasses may wish to override the ``to_slack`` method.

    Parameters
    ----------
    message
        Exception string value, which is the default Slack message.
    failed_at
        When the exception happened. Omit to use the current time.
    method
        Method of request.
    url
        URL of the request.
    user
        Username on whose behalf the request is being made.
    status
        Status code of failure, if any.
    body
        Body of failure message, if any.
    """

    @classmethod
    def from_exception(cls, exc: HTTPError, user: str | None = None) -> Self:
        """Create an exception from an HTTPX_ exception.

        Parameters
        ----------
        exc
            Exception from HTTPX.
        user
            User on whose behalf the request is being made, if known.

        Returns
        -------
        SlackWebException
            Newly-constructed exception.
        """
        if isinstance(exc, HTTPStatusError):
            status = exc.response.status_code
            method = exc.request.method
            message = f"Status {status} from {method} {exc.request.url}"
            return cls(
                message,
                method=exc.request.method,
                url=str(exc.request.url),
                user=user,
                status=status,
                body=exc.response.text,
            )
        else:
            exc_name = type(exc).__name__
            message = f"{exc_name}: {exc!s}" if str(exc) else exc_name

            # All httpx.HTTPError exceptions have a slot for the request,
            # initialized to None and then sometimes added by child
            # constructors or during exception processing. The request
            # property is a property method that raises RuntimeError if
            # request has not been set, so we can't just check for None. Hence
            # this approach of attempting to use the request and falling back
            # on reporting less data if that raised any exception.
            try:
                return cls(
                    message,
                    method=exc.request.method,
                    url=str(exc.request.url),
                    user=user,
                )
            except Exception:
                return cls(message, user=user)

    def __init__(
        self,
        message: str,
        *,
        failed_at: datetime | None = None,
        method: str | None = None,
        url: str | None = None,
        user: str | None = None,
        status: int | None = None,
        body: str | None = None,
    ) -> None:
        super().__init__(message, user, failed_at=failed_at)
        self.method = method
        self.url = url
        self.status = status
        self.body = body

    def __str__(self) -> str:
        result = self.message
        if self.body:
            result += f"\nBody:\n{self.body}\n"
        return result

    @override
    def to_slack(self) -> SlackMessage:
        """Convert to a Slack message for Slack alerting.

        Returns
        -------
        SlackMessage
            Slack message suitable for posting as an alert.
        """
        message = super().to_slack()
        message.message = self.message
        if self.url:
            text = f"{self.method} {self.url}" if self.method else self.url
            message.blocks.append(SlackTextBlock(heading="URL", text=text))
        if self.body:
            block = SlackCodeBlock(heading="Response", code=self.body)
            message.attachments.append(block)
        return message

    @override
    def to_sentry(self) -> SentryEventInfo:
        """Convert to Sentry attributes for Sentry encrichment.

        Returns
        -------
        SentryEventInfo
            Sentry attributes to enrich an event generated from this exception.
        """
        info = super().to_sentry()
        if self.url:
            if self.method:
                info.tags["httpx_request_method"] = self.method
            if self.url:
                info.tags["httpx_request_url"] = self.url
        if self.body:
            info.attachments["httpx_response_body"] = self.body
        return info


def _format_and_truncate_at_end(string: str, max_length: int) -> str:
    """Format a string for Slack, truncating at the end.

    Slack prohibits text blocks longer than a varying number of characters
    depending on where they are in the message. If this constraint is not met,
    the whole mesage is rejected with an HTTP error. Truncate a potentially
    long message at the end.

    Parameters
    ----------
    string
        String to truncate.
    max_length
        Maximum allowed length.

    Returns
    -------
    str
        The truncated string with special characters escaped.
    """
    string = (
        string.strip()
        .replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )
    if len(string) <= max_length:
        return string
    truncated = "\n... truncated ..."
    last_newline = string.rfind("\n", 0, max_length - len(truncated))
    if last_newline == -1:
        return string[: max_length - len(truncated)] + truncated
    else:
        return string[:last_newline] + truncated


def _format_and_truncate_at_start(string: str, max_length: int) -> str:
    """Format a string for Slack, truncating at the start.

    Slack prohibits text blocks longer than a varying number of characters
    depending on where they are in the message. If this constraint is not met,
    the whole mesage is rejected with an HTTP error. Truncate a potentially
    long message at the start. Use this for tracebacks and similar

    Parameters
    ----------
    string
        String to truncate.
    max_length
        Maximum allowed length.

    Returns
    -------
    str
        The truncated string with special characters escaped.
    """
    string = (
        string.strip()
        .replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )
    length = len(string)
    if length <= max_length:
        return string
    truncated = "... truncated ...\n"
    lines = string.split("\n")
    if len(lines) == 1:
        start = length - max_length + len(truncated)
        return truncated + string[start:]
    while length > max_length - len(truncated):
        line = lines.pop(0)
        length -= len(line) + 1
    return truncated + "\n".join(lines)



"""Constants for test containers."""

CONFLUENT_VERSION_TAG = "7.6.0"
"""Docker image tag for Confluent images (Kafka and schema registry)."""

KAFKA_REPOSITORY = "confluentinc/cp-kafka"
"""Docker repository for Kafka."""

SCHEMA_REGISTRY_REPOSITORY = "confluentinc/cp-schema-registry"
"""Docker repository for the Confluent schema registry."""



"""Utilities for configuring uvicorn to use structlog."""

from __future__ import annotations

import logging
import logging.config
import re

import structlog
from structlog.types import EventDict

from ._models import LogLevel
from ._structlog import add_log_severity

__all__ = ["configure_uvicorn_logging"]

_UVICORN_ACCESS_REGEX = re.compile(r'^[0-9.]+:[0-9]+ - "([^"]+)" ([0-9]+)$')
"""Regex to parse Uvicorn access logs."""


def _process_uvicorn_access_log(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """Parse a Uvicorn access log entry into key/value pairs.

    Intended for use as a structlog processor.

    This checks whether the log message is a Uvicorn access log entry and, if
    so, parses the message into key/value pairs for JSON logging so that the
    details can be programmatically extracted.  ``remoteIp`` is intentionally
    omitted since it isn't aware of ``X-Forwarded-For`` and will therefore
    always point to an uninteresting in-cluster IP.

    Parameters
    ----------
    logger
        The wrapped logger object.
    method_name
        The name of the wrapped method (``warning`` or ``error``, for
        example).
    event_dict
        Current context and current event. This parameter is also modified in
        place, matching the normal behavior of structlog processors.

    Returns
    -------
    EventDict
        The modified `~structlog.types.EventDict` with the added key.
    """
    match = _UVICORN_ACCESS_REGEX.match(event_dict["event"])
    if not match:
        return event_dict
    request = match.group(1)
    method, rest = request.split(" ", 1)
    url, protocol = rest.rsplit(" ", 1)
    if "httpRequest" not in event_dict:
        event_dict["httpRequest"] = {}
    event_dict["httpRequest"]["protocol"] = protocol
    event_dict["httpRequest"]["requestMethod"] = method
    event_dict["httpRequest"]["requestUrl"] = url
    event_dict["httpRequest"]["status"] = match.group(2)
    return event_dict


def configure_uvicorn_logging(
    log_level: LogLevel | str = LogLevel.INFO,
) -> None:
    """Set up logging.

    This configures Uvicorn to use structlog for output formatting and
    installs a custom processor to parse its access log messages into
    additional log context that matches the format of Google log messages.
    This helps Google's Cloud Logging system understand the logs. Access logs
    are sent to standard output and all other logs are sent to standard
    error.

    Parameters
    ----------
    log_level
        The Python log level. May be given as a `LogLevel` enum (preferred)
        or a case-insensitive string.

    Notes
    -----
    This method should normally be called during FastAPI app creation, either
    during Python module import or inside the function that creates and
    returns the FastAPI app that Uvicorn will run. This ensures the logging
    setup is complete before Uvicorn logs its first message.
    """
    if not isinstance(log_level, LogLevel):
        log_level = LogLevel[log_level.upper()]

    processors = [
        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
        structlog.processors.format_exc_info,
        structlog.processors.JSONRenderer(),
    ]
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "json-access": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": processors,
                    "foreign_pre_chain": [
                        add_log_severity,
                        _process_uvicorn_access_log,
                    ],
                },
                "json": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": processors,
                    "foreign_pre_chain": [add_log_severity],
                },
            },
            "handlers": {
                "uvicorn.access": {
                    "level": log_level.value,
                    "class": "logging.StreamHandler",
                    "formatter": "json-access",
                    "stream": "ext://sys.stdout",
                },
                "uvicorn.default": {
                    "level": log_level.value,
                    "class": "logging.StreamHandler",
                    "formatter": "json",
                },
            },
            "loggers": {
                "uvicorn.error": {
                    "handlers": ["uvicorn.default"],
                    "level": log_level.value,
                    "propagate": False,
                },
                "uvicorn.access": {
                    "handlers": ["uvicorn.access"],
                    "level": log_level.value,
                    "propagate": False,
                },
            },
        }
    )



"""HTTP client dependency for FastAPI."""

from __future__ import annotations

import httpx

__all__ = [
    "DEFAULT_HTTP_TIMEOUT",
    "HTTPClientDependency",
    "http_client_dependency",
]

DEFAULT_HTTP_TIMEOUT = 20.0
"""Default timeout (in seconds) for outbound HTTP requests.

The default HTTPX timeout has proven too short in practice for calls to, for
example, GitHub for authorization verification. Increase the default to 20
seconds. Users of this dependency can always lower it if needed.
"""


class HTTPClientDependency:
    """Provides an ``httpx.AsyncClient`` as a dependency.

    The resulting client will have redirects enabled and the default timeout
    increased to 20 seconds.

    Notes
    -----
    The application must call ``http_client_dependency.aclose()`` in the
    application lifespan hook:

    .. code-block:: python

       from collections.abc import AsyncGenerator
       from contextlib import asynccontextmanager

       from fastapi import FastAPI


       @asynccontextmanager
       async def lifespan(app: FastAPI) -> AsyncGenerator[None]:
           yield
           await http_client_dependency.aclose()


       app = FastAPI(lifespan=lifespan)
    """

    def __init__(self) -> None:
        self.http_client: httpx.AsyncClient | None = None

    async def __call__(self) -> httpx.AsyncClient:
        """Return the cached ``httpx.AsyncClient``."""
        if not self.http_client:
            self.http_client = httpx.AsyncClient(
                timeout=DEFAULT_HTTP_TIMEOUT, follow_redirects=True
            )
        return self.http_client

    async def aclose(self) -> None:
        """Close the ``httpx.AsyncClient``."""
        if self.http_client:
            await self.http_client.aclose()
            self.http_client = None


http_client_dependency = HTTPClientDependency()
"""The dependency that will return the HTTP client."""



"""Constants for use in tests."""

CONFLUENT_VERSION_TAG = "7.6.0"
"""Docker image tag for Confluent images like kafka and the schema registry."""

KAFKA_DOCKER_IMAGE = f"confluentinc/cp-kafka:{CONFLUENT_VERSION_TAG}"
"""Docker image for Kafka Testcontainer"""

SCHEMA_REGISTRY_DOCKER_IMAGE = (
    f"confluentinc/cp-schema-registry:{CONFLUENT_VERSION_TAG}"
)
"""Docker image for schema registry Testcontainer"""



"""Support library for writing UWS-enabled services."""

from ._app import UWSApplication
from ._config import UWSAppSettings, UWSConfig, UWSRoute
from ._exceptions import ParameterError, UsageError, UWSError
from ._models import (
    Job,
    JobBase,
    JobCreate,
    JobError,
    JobResult,
    JobUpdateAborted,
    JobUpdateCompleted,
    JobUpdateError,
    JobUpdateExecuting,
    JobUpdateMetadata,
    JobUpdateQueued,
    ParametersModel,
    SerializedJob,
)

__all__ = [
    "Job",
    "JobBase",
    "JobCreate",
    "JobError",
    "JobResult",
    "JobUpdateAborted",
    "JobUpdateCompleted",
    "JobUpdateError",
    "JobUpdateExecuting",
    "JobUpdateMetadata",
    "JobUpdateQueued",
    "ParameterError",
    "ParametersModel",
    "SerializedJob",
    "UWSAppSettings",
    "UWSApplication",
    "UWSConfig",
    "UWSError",
    "UWSRoute",
    "UsageError",
]



"""An arq_ client with a mock for testing."""

from __future__ import annotations

import abc
import asyncio
import uuid
from datetime import UTC, datetime
from typing import Any, Self, override

from arq import create_pool
from arq.connections import ArqRedis, RedisSettings
from arq.constants import default_queue_name as arq_default_queue_name
from arq.jobs import Job, JobStatus

from ._exceptions import JobNotFound, JobNotQueued, JobResultUnavailable
from ._models import JobMetadata, JobResult

__all__ = [
    "ArqQueue",
    "MockArqQueue",
    "RedisArqQueue",
]


class ArqQueue(metaclass=abc.ABCMeta):
    """A common interface for working with an arq queue that can be
    implemented either with a real Redis backend, or an in-memory repository
    for testing.

    See Also
    --------
    RedisArqQueue
        Production implementation with a Redis store.
    MockArqQueue
        In-memory implementation for testing and development.
    """

    def __init__(
        self, *, default_queue_name: str = arq_default_queue_name
    ) -> None:
        self._default_queue_name = default_queue_name

    @property
    def default_queue_name(self) -> str:
        """Name of the default queue, if the ``_queue_name`` parameter is not
        set in method calls.
        """
        return self._default_queue_name

    @abc.abstractmethod
    async def enqueue(
        self,
        task_name: str,
        *task_args: Any,
        _queue_name: str | None = None,
        **task_kwargs: Any,
    ) -> JobMetadata:
        """Add a job to the queue.

        Parameters
        ----------
        task_name
            The function name to run.
        *task_args
            Positional arguments for the task function.
        _queue_name
            Name of the queue.
        **task_kwargs
            Keyword arguments passed to the task function.

        Returns
        -------
        JobMetadata
            Metadata about the queued job.

        Raises
        ------
        JobNotQueued
            Raised if the job is not successfully added to the queue.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def abort_job(
        self,
        job_id: str,
        queue_name: str | None = None,
        *,
        timeout: float | None = None,
    ) -> bool:
        """Abort a queued or running job.

        The worker must be configured to allow aborting jobs for this to
        succeed.

        Parameters
        ----------
        job_id
            The job's identifier.
        queue_name
            Name of the queue.
        timeout
            How long to wait for the job result before raising `TimeoutError`.
            If `None`, waits forever.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def get_job_metadata(
        self, job_id: str, queue_name: str | None = None
    ) -> JobMetadata:
        """Get metadata about a `~arq.jobs.Job`.

        Parameters
        ----------
        job_id
            The job's identifier. This is the same as the `JobMetadata.id`
            attribute, provided when initially adding a job to the queue.
        queue_name
            Name of the queue.

        Returns
        -------
        JobMetadata
            Metadata about the queued job.

        Raises
        ------
        JobNotFound
            Raised if the job is not found in the queue.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def get_job_result(
        self, job_id: str, queue_name: str | None = None
    ) -> JobResult:
        """Retrieve the job result, if available.

        Parameters
        ----------
        job_id
            The job's identifier. This is the same as the `JobMetadata.id`
            attribute, provided when initially adding a job to the queue.
        queue_name
            Name of the queue.

        Returns
        -------
        JobResult
            The job's result, along with metadata about the queued job.

        Raises
        ------
        JobNotFound
            Raised if the job is not found in the queue.
        JobResultUnavailable
            Raised if the job's result is unavailable for any reason.
        """
        raise NotImplementedError


class RedisArqQueue(ArqQueue):
    """A distributed queue, based on arq and Redis."""

    def __init__(
        self,
        pool: ArqRedis,
        *,
        default_queue_name: str = arq_default_queue_name,
    ) -> None:
        super().__init__(default_queue_name=default_queue_name)
        self._pool = pool

    @classmethod
    async def initialize(
        cls,
        redis_settings: RedisSettings,
        *,
        default_queue_name: str = arq_default_queue_name,
    ) -> Self:
        """Initialize a RedisArqQueue from Redis settings."""
        pool = await create_pool(
            redis_settings, default_queue_name=default_queue_name
        )
        return cls(pool, default_queue_name=default_queue_name)

    @override
    async def enqueue(
        self,
        task_name: str,
        *task_args: Any,
        _queue_name: str | None = None,
        **task_kwargs: Any,
    ) -> JobMetadata:
        job = await self._pool.enqueue_job(
            task_name,
            *task_args,
            _queue_name=_queue_name or self.default_queue_name,
            **task_kwargs,
        )
        if job:
            return await JobMetadata.from_job(job)
        else:
            # TODO(jonathansick): if implementing hard-coded job IDs, set as
            # an argument
            raise JobNotQueued(None)

    def _get_job(self, job_id: str, queue_name: str | None = None) -> Job:
        return Job(
            job_id,
            self._pool,
            _queue_name=queue_name or self.default_queue_name,
        )

    @override
    async def abort_job(
        self,
        job_id: str,
        queue_name: str | None = None,
        *,
        timeout: float | None = None,
    ) -> bool:
        job = self._get_job(job_id, queue_name=queue_name)
        return await job.abort(timeout=timeout)

    @override
    async def get_job_metadata(
        self, job_id: str, queue_name: str | None = None
    ) -> JobMetadata:
        job = self._get_job(job_id, queue_name=queue_name)
        return await JobMetadata.from_job(job)

    @override
    async def get_job_result(
        self, job_id: str, queue_name: str | None = None
    ) -> JobResult:
        job = self._get_job(job_id, queue_name=queue_name)
        return await JobResult.from_job(job)


class MockArqQueue(ArqQueue):
    """A mocked queue for testing API services."""

    def __init__(
        self, *, default_queue_name: str = arq_default_queue_name
    ) -> None:
        super().__init__(default_queue_name=default_queue_name)
        self._job_metadata: dict[str, dict[str, JobMetadata]] = {
            self.default_queue_name: {}
        }
        self._job_results: dict[str, dict[str, JobResult]] = {
            self.default_queue_name: {}
        }

    def _resolve_queue_name(self, queue_name: str | None) -> str:
        return queue_name or self.default_queue_name

    @override
    async def enqueue(
        self,
        task_name: str,
        *task_args: Any,
        _queue_name: str | None = None,
        **task_kwargs: Any,
    ) -> JobMetadata:
        queue_name = self._resolve_queue_name(_queue_name)
        new_job = JobMetadata(
            id=str(uuid.uuid4().hex),
            name=task_name,
            args=task_args,
            kwargs=task_kwargs,
            enqueue_time=datetime.now(tz=UTC),
            status=JobStatus.queued,
            queue_name=queue_name,
        )
        self._job_metadata[queue_name][new_job.id] = new_job
        return new_job

    @override
    async def abort_job(
        self,
        job_id: str,
        queue_name: str | None = None,
        *,
        timeout: float | None = None,
    ) -> bool:
        queue_name = self._resolve_queue_name(queue_name)
        try:
            job_metadata = self._job_metadata[queue_name][job_id]
        except KeyError:
            return False

        # If the job was started, simulate cancelling it.
        if job_metadata.status == JobStatus.in_progress:
            job_metadata.status = JobStatus.complete
            result_info = JobResult(
                id=job_metadata.id,
                name=job_metadata.name,
                args=job_metadata.args,
                kwargs=job_metadata.kwargs,
                status=job_metadata.status,
                enqueue_time=job_metadata.enqueue_time,
                start_time=datetime.now(tz=UTC),
                finish_time=datetime.now(tz=UTC),
                result=asyncio.CancelledError(),
                success=False,
                queue_name=queue_name,
            )
            self._job_results[queue_name][job_id] = result_info
            return True

        # If it was just queued, delete it.
        if job_metadata.status in (JobStatus.deferred, JobStatus.queued):
            del self._job_metadata[queue_name][job_id]
            return True

        # Otherwise, the job has already completed, so we can't abort it.
        return False

    @override
    async def get_job_metadata(
        self, job_id: str, queue_name: str | None = None
    ) -> JobMetadata:
        queue_name = self._resolve_queue_name(queue_name)
        try:
            return self._job_metadata[queue_name][job_id]
        except KeyError as e:
            raise JobNotFound(job_id) from e

    @override
    async def get_job_result(
        self, job_id: str, queue_name: str | None = None
    ) -> JobResult:
        queue_name = self._resolve_queue_name(queue_name)
        try:
            return self._job_results[queue_name][job_id]
        except KeyError as e:
            raise JobResultUnavailable(job_id) from e

    async def set_in_progress(
        self, job_id: str, queue_name: str | None = None
    ) -> None:
        """Set a job's status to in progress, for mocking a queue in tests."""
        job = await self.get_job_metadata(job_id, queue_name=queue_name)
        job.status = JobStatus.in_progress

        # An in-progress job cannot have a result
        if job_id in self._job_results:
            del self._job_results[job_id]

    async def set_complete(
        self,
        job_id: str,
        *,
        result: Any,
        success: bool = True,
        queue_name: str | None = None,
    ) -> None:
        """Set a job's result, for mocking a queue in tests."""
        queue_name = self._resolve_queue_name(queue_name)

        job_metadata = await self.get_job_metadata(
            job_id, queue_name=queue_name
        )
        job_metadata.status = JobStatus.complete

        result_info = JobResult(
            id=job_metadata.id,
            name=job_metadata.name,
            args=job_metadata.args,
            kwargs=job_metadata.kwargs,
            status=job_metadata.status,
            enqueue_time=job_metadata.enqueue_time,
            start_time=datetime.now(tz=UTC),
            finish_time=datetime.now(tz=UTC),
            result=result,
            success=success,
            queue_name=queue_name,
        )
        self._job_results[queue_name][job_id] = result_info



"""A decorator to run a function under `asyncio.run`."""

import asyncio
from collections.abc import Callable, Coroutine
from functools import wraps

__all__ = ["run_with_asyncio"]


def run_with_asyncio[**P, F](
    f: Callable[P, Coroutine[None, None, F]],
) -> Callable[P, F]:
    """Run the decorated function with `asyncio.run`.

    Intended to be used as a decorator around an async function that needs to
    be run in a sync context.  The decorated function will be run with
    `asyncio.run` when invoked.  The caller must not already be inside an
    asyncio task.

    Parameters
    ----------
    f
        The function to wrap.

    Examples
    --------
    An application that uses Safir and `Click`_ may use the following Click
    command function to initialize a database.

    .. code-block:: python

       import structlog
       from safir.asyncio import run_with_asyncio
       from safir.database import initialize_database

       from .config import config
       from .schema import Base


       @main.command()
       @run_with_asyncio
       async def init() -> None:
           logger = structlog.get_logger(config.safir.logger_name)
           engine = await initialize_database(
               config.database_url,
               config.database_password,
               logger,
               schema=Base.metadata,
           )
           await engine.dispose()
    """

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> F:
        return asyncio.run(f(*args, **kwargs))

    return wrapper



"""Tests for the database utility functions."""

from __future__ import annotations

import asyncio
import os
import subprocess
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta, timezone
from pathlib import Path
from typing import Any, Self
from urllib.parse import unquote, urlparse

import pytest
import structlog
from pydantic import BaseModel, SecretStr
from pydantic_core import Url
from sqlalchemy import Column, MetaData, Select, String, Table, select
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.orm import (
    DeclarativeBase,
    InstrumentedAttribute,
    Mapped,
    mapped_column,
)
from starlette.datastructures import URL

from safir.database import (
    CountedPaginatedQueryRunner,
    DatetimeIdCursor,
    PaginatedQueryRunner,
    PaginationLinkData,
    create_async_session,
    create_database_engine,
    datetime_from_db,
    datetime_to_db,
    drop_database,
    initialize_database,
    is_database_current,
    retry_async_transaction,
    stamp_database,
    stamp_database_async,
    unstamp_database,
)
from safir.database._connection import build_database_url
from safir.pydantic import UtcDatetime

from .support.alembic import BaseV1, BaseV2, UserV1, UserV2, config


@pytest.mark.asyncio
async def test_initialize_database(
    database_url: str, database_password: str
) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )
    session = await create_async_session(engine, logger)
    async with session.begin():
        session.add(UserV2(username="someuser"))
    await session.remove()

    # Reinitializing the database without reset should preserve the row.
    await initialize_database(engine, logger, schema=BaseV2.metadata)
    session = await create_async_session(engine, logger)
    async with session.begin():
        result = await session.scalars(select(UserV2.username))
        assert result.all() == ["someuser"]
    await session.remove()

    # Reinitializing the database with reset should delete the data. Try
    # passing in the password as a SecretStr.
    password = SecretStr(database_password)
    engine = create_database_engine(database_url, password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )
    session = await create_async_session(engine, logger)
    async with session.begin():
        result = await session.scalars(select(UserV2.username))
        assert result.all() == []
    await session.remove()
    await engine.dispose()


@pytest.mark.asyncio
async def test_drop_database(
    database_url: str, database_password: str
) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(engine, logger, schema=BaseV2.metadata)
    session = await create_async_session(engine, logger)
    async with session.begin():
        session.add(UserV2(username="someuser"))
    await session.remove()

    await drop_database(engine, BaseV2.metadata)
    session = await create_async_session(engine, logger)
    with pytest.raises(ProgrammingError):
        async with session.begin():
            await session.scalars(select(UserV2.username))
    await session.remove()
    await engine.dispose()


def test_build_database_url(database_url: str) -> None:
    url = build_database_url("postgresql://foo@127.0.0.1/foo", None)
    assert url == "postgresql+asyncpg://foo@127.0.0.1/foo"

    url = build_database_url("postgresql://foo@127.0.0.1:5432/foo", None)
    assert url == "postgresql+asyncpg://foo@127.0.0.1:5432/foo"

    url = build_database_url("postgresql://foo@127.0.0.1/foo", "otherpass")
    assert url == "postgresql+asyncpg://foo:otherpass@127.0.0.1/foo"

    url = build_database_url(
        "postgresql://foo@127.0.0.1:5433/foo", "otherpass"
    )
    assert url == "postgresql+asyncpg://foo:otherpass@127.0.0.1:5433/foo"

    pydantic_url = Url.build(
        scheme="postgresql", username="user", host="localhost", path="foo"
    )
    url = build_database_url(pydantic_url, "password")
    assert url == "postgresql+asyncpg://user:password@localhost/foo"

    # Test that the username and password are quoted properly.
    url = build_database_url(
        "postgresql://foo%40e.com@127.0.0.1:4444/foo",
        "pass@word/with stuff",
    )
    assert url == (
        "postgresql+asyncpg://foo%40e.com:pass%40word%2Fwith%20stuff"
        "@127.0.0.1:4444/foo"
    )
    parsed_url = urlparse(url)
    assert parsed_url.username
    assert parsed_url.password

    # urlparse does not undo quoting in the components of netloc.
    assert unquote(parsed_url.username) == "foo@e.com"
    assert unquote(parsed_url.password) == "pass@word/with stuff"
    assert parsed_url.hostname == "127.0.0.1"
    assert parsed_url.port == 4444


@pytest.mark.asyncio
async def test_create_async_session(
    database_url: str, database_password: str
) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )

    session = await create_async_session(
        engine, logger, statement=select(UserV2)
    )
    async with session.begin():
        session.add(UserV2(username="foo"))
    await session.remove()

    # Use a query against a non-existent table as the liveness check and
    # ensure that fails.
    metadata = MetaData()
    bad_table = Table("bad", metadata, Column("name", String(64)))
    with pytest.raises(ProgrammingError):
        session = await create_async_session(
            engine, logger, statement=select(bad_table)
        )
    await engine.dispose()


def test_datetime() -> None:
    tz_aware = datetime.now(tz=UTC)
    tz_naive = tz_aware.replace(tzinfo=None)

    assert datetime_to_db(tz_aware) == tz_naive
    assert datetime_from_db(tz_naive) == tz_aware
    assert datetime_from_db(tz_aware) == tz_aware

    assert datetime_to_db(None) is None
    assert datetime_from_db(None) is None

    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_to_db(tz_naive)

    tz_local = datetime.now(tz=timezone(timedelta(hours=1)))
    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_to_db(tz_local)
    with pytest.raises(ValueError, match=r"datetime .* not in UTC"):
        datetime_from_db(tz_local)

    # Pydantic's JSON decoder uses a TzInfo data structure instead of
    # datetime.timezone.utc. Make sure that's still recognized as UTC.
    class Test(BaseModel):
        time: datetime

    json_model = Test(time=tz_aware).model_dump_json()
    model = Test.model_validate_json(json_model)
    assert datetime_to_db(model.time) == tz_naive


@pytest.mark.asyncio
async def test_retry_async_transaction(
    database_url: str, database_password: str
) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(
        engine, logger, schema=BaseV2.metadata, reset=True
    )
    session = await create_async_session(engine, logger)
    async with session.begin():
        session.add(UserV2(username="someuser"))
    tries = 0

    @retry_async_transaction
    async def insert(attempts: int) -> None:
        nonlocal tries
        tries += 1
        async with session.begin():
            if tries <= attempts:
                raise OperationalError(None, None, ValueError("foo"))
            session.add(UserV2(username="newuser"))

    await insert(2)

    tries = 0
    with pytest.raises(OperationalError):
        await insert(3)

    @retry_async_transaction(max_tries=1)
    async def insert_capped(attempts: int) -> None:
        nonlocal tries
        tries += 1
        async with session.begin():
            if tries <= attempts:
                raise OperationalError(None, None, ValueError("foo"))
            session.add(UserV2(username="newuser"))

    tries = 0
    with pytest.raises(OperationalError):
        await insert_capped(1)

    await session.remove()
    await engine.dispose()


def test_alembic(
    database_url: str,
    database_password: str,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    event_loop = asyncio.new_event_loop()
    config.database_url = Url(database_url)
    config.database_password = SecretStr(database_password)
    config_path = (
        Path(__file__).parent / "data" / "database" / "v1" / "alembic.ini"
    )
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)

    async def init(schema: MetaData) -> None:
        await initialize_database(engine, logger, schema=schema, reset=True)

    async def check(config_path: Path) -> bool:
        return await is_database_current(engine, logger, config_path)

    async def store(record: Any) -> None:
        session = await create_async_session(engine, logger)
        async with session.begin():
            session.add(record)
        await session.remove()

    async def list_v2() -> list[str]:
        session = await create_async_session(engine, logger)
        async with session.begin():
            result = await session.scalars(select(UserV2.username))
        await session.remove()
        return list(result.all())

    # Initialize the database with the V1 schema.
    event_loop.run_until_complete(init(BaseV1.metadata))

    # Before stamping, the database should show as not up to date.
    assert not event_loop.run_until_complete(check(config_path))

    # Stamp the database.
    stamp_database(config_path)

    # Check that the database is up to date.
    assert event_loop.run_until_complete(check(config_path))

    # Store a record in the database.
    event_loop.run_until_complete(store(UserV1(username="foo", name="Foo")))

    # Check that the database is not current with the V2 schema.
    config_path = config_path.parent.parent / "v2" / "alembic.ini"
    assert not event_loop.run_until_complete(check(config_path))

    # Upgrade the database to V2. It should now show as current with the V2
    # schema.
    subprocess.run(
        ["alembic", "upgrade", "head"],
        check=True,
        cwd=str(config_path.parent),
        env={
            **os.environ,
            "TEST_DATABASE_URL": database_url,
            "TEST_DATABASE_PASSWORD": database_password,
            "PYTHONPATH": str(Path(__file__).parent.parent),
        },
    )
    assert event_loop.run_until_complete(check(config_path))

    # The data should still be there.
    assert event_loop.run_until_complete(list_v2()) == ["foo"]

    # Unstamping the database should cause it to no longer be current, but we
    # should be able to them stamp it async and then it should be current
    # again.
    event_loop.run_until_complete(unstamp_database(engine))
    assert not event_loop.run_until_complete(check(config_path))
    event_loop.run_until_complete(stamp_database_async(engine, config_path))
    assert event_loop.run_until_complete(check(config_path))


class PaginationBase(DeclarativeBase):
    pass


class PaginationTable(PaginationBase):
    __tablename__ = "table"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    time: Mapped[datetime]

    def __repr__(self) -> str:
        return f"PaginationTable(id={self.id}, time={self.time})"


class PaginationModel(BaseModel):
    id: int
    time: UtcDatetime


@dataclass
class TableCursor(DatetimeIdCursor[PaginationModel]):
    @staticmethod
    def id_column() -> InstrumentedAttribute:
        return PaginationTable.id

    @staticmethod
    def time_column() -> InstrumentedAttribute:
        return PaginationTable.time

    @classmethod
    def from_entry(
        cls, entry: PaginationModel, *, reverse: bool = False
    ) -> Self:
        return cls(time=entry.time, id=entry.id, previous=reverse)


def naive_datetime(timestamp: float) -> datetime:
    """Construct timezone-naive datetimes for tests."""
    return datetime_to_db(datetime.fromtimestamp(timestamp, tz=UTC))


def assert_model_lists_equal(
    a: list[PaginationModel], b: list[PaginationTable]
) -> None:
    assert len(a) == len(b)
    for index, entry in enumerate(a):
        assert entry.id == b[index].id, f"element {index} id"
        orm_time = b[index].time.replace(tzinfo=UTC)
        assert entry.time == orm_time, f"element {index} time"


@pytest.mark.asyncio
async def test_pagination(database_url: str, database_password: str) -> None:
    logger = structlog.get_logger(__name__)
    engine = create_database_engine(database_url, database_password)
    await initialize_database(
        engine, logger, schema=PaginationBase.metadata, reset=True
    )
    session = await create_async_session(engine, logger)

    # Add the test database rows in a random order so that the ordering of
    # unique IDs does not match the ordering of the timestamps.
    rows = [
        PaginationTable(time=naive_datetime(1600000000.5)),
        PaginationTable(time=naive_datetime(1510000000)),
        PaginationTable(time=naive_datetime(1520000000)),
        PaginationTable(time=naive_datetime(1500000000)),
        PaginationTable(time=naive_datetime(1520000000)),
        PaginationTable(time=naive_datetime(1600000000.5)),
        PaginationTable(time=naive_datetime(1610000000)),
    ]
    async with session.begin():
        for row in rows:
            session.add(row)

    # Rows will be returned from the database in reverse order, so change the
    # rows data structure to match.
    rows.sort(key=lambda e: (e.time, e.id), reverse=True)

    # Query by object and test the pagination cursors going backwards and
    # forwards.
    runner = PaginatedQueryRunner(PaginationModel, TableCursor)
    counted_runner = CountedPaginatedQueryRunner(PaginationModel, TableCursor)
    async with session.begin():
        stmt: Select[tuple] = select(PaginationTable)
        assert await runner.query_count(session, stmt) == 7
        result = await runner.query_object(session, stmt, limit=2)
        assert_model_lists_equal(result.entries, rows[:2])
        assert not result.prev_cursor
        base_url = URL("https://example.com/query")
        next_url = f"{base_url!s}?cursor={result.next_cursor}"
        assert result.link_header(base_url) == (
            f'<{base_url!s}>; rel="first", <{next_url}>; rel="next"'
        )
        assert result.first_url(base_url) == str(base_url)
        assert result.next_url(base_url) == next_url
        assert result.prev_url(base_url) is None
        assert str(result.next_cursor) == "1600000000.5_1"

        counted_result = await counted_runner.query_object(
            session, stmt, limit=2
        )
        assert counted_result.entries == result.entries
        assert counted_result.prev_cursor == result.prev_cursor
        assert counted_result.next_cursor == result.next_cursor
        assert counted_result.count == 7

        result = await runner.query_object(
            session, stmt, cursor=result.next_cursor, limit=3
        )
        assert_model_lists_equal(result.entries, rows[2:5])
        assert str(result.next_cursor) == "1510000000_2"
        assert str(result.prev_cursor) == "p1600000000.5_1"
        base_url = URL("https://example.com/query?foo=bar&foo=baz&cursor=xxxx")
        stripped_url = "https://example.com/query?foo=bar&foo=baz"
        next_url = f"{stripped_url}&cursor={result.next_cursor}"
        prev_url = f"{stripped_url}&cursor={result.prev_cursor}"
        assert result.link_header(base_url) == (
            f'<{stripped_url}>; rel="first", '
            f'<{next_url}>; rel="next", '
            f'<{prev_url}>; rel="prev"'
        )
        assert result.first_url(base_url) == stripped_url
        assert result.next_url(base_url) == next_url
        assert result.prev_url(base_url) == prev_url
        next_cursor = result.next_cursor

        result = await runner.query_object(
            session, stmt, cursor=result.prev_cursor
        )
        assert_model_lists_equal(result.entries, rows[:2])
        base_url = URL("https://example.com/query?limit=2")
        assert result.link_header(base_url) == (
            f'<{base_url!s}>; rel="first", '
            f'<{base_url!s}&cursor={result.next_cursor}>; rel="next"'
        )

        result = await runner.query_object(session, stmt, cursor=next_cursor)
        assert_model_lists_equal(result.entries, rows[5:])
        assert not result.next_cursor
        base_url = URL("https://example.com/query")
        assert result.next_url(base_url) is None
        assert result.link_header(base_url) == (
            f'<{base_url!s}>; rel="first", '
            f'<{base_url!s}?cursor={result.prev_cursor}>; rel="prev"'
        )
        prev_cursor = result.prev_cursor

        result = await runner.query_object(session, stmt, cursor=prev_cursor)
        assert_model_lists_equal(result.entries, rows[:5])
        assert result.link_header(base_url) == (
            f'<{base_url!s}>; rel="first", '
            f'<{base_url!s}?cursor={result.next_cursor}>; rel="next"'
        )

        result = await runner.query_object(
            session, stmt, cursor=prev_cursor, limit=2
        )
        assert_model_lists_equal(result.entries, rows[3:5])
        assert str(result.prev_cursor) == "p1520000000_5"
        assert result.link_header(base_url) == (
            f'<{base_url!s}>; rel="first", '
            f'<{base_url!s}?cursor={result.next_cursor}>; rel="next", '
            f'<{base_url!s}?cursor={result.prev_cursor}>; rel="prev"'
        )

    # Perform one of the queries by attribute instead to test the query_row
    # function.
    async with session.begin():
        stmt = select(PaginationTable.time, PaginationTable.id)
        result = await runner.query_row(session, stmt, limit=2)
        assert_model_lists_equal(result.entries, rows[:2])
        assert await runner.query_count(session, stmt) == 7

        counted_result = await counted_runner.query_row(session, stmt, limit=2)
        assert counted_result.entries == result.entries
        assert counted_result.prev_cursor == result.prev_cursor
        assert counted_result.next_cursor == result.next_cursor
        assert counted_result.count == 7

    # Querying for the entire table should return the everything with no
    # pagination cursors. Try this with both an object query and an attribute
    # query.
    async with session.begin():
        stmt = select(PaginationTable)
        result = await runner.query_object(session, stmt)
        assert_model_lists_equal(result.entries, rows)
        assert not result.next_cursor
        assert not result.prev_cursor
        stmt = select(PaginationTable.id, PaginationTable.time)
        result = await runner.query_row(session, stmt)
        assert_model_lists_equal(result.entries, rows)
        assert not result.next_cursor
        assert not result.prev_cursor
        base_url = URL("https://example.com/query?foo=b")
        assert result.link_header(base_url) == (f'<{base_url!s}>; rel="first"')

        counted_result = await counted_runner.query_row(session, stmt)
        assert counted_result.entries == result.entries
        assert not counted_result.next_cursor
        assert not counted_result.prev_cursor
        assert counted_result.count == len(counted_result.entries)


def test_link_data() -> None:
    header = (
        '<https://example.com/query>; rel="first", '
        '<https://example.com/query?cursor=1600000000.5_1>; rel="next"'
    )
    link = PaginationLinkData.from_header(header)
    assert not link.prev_url
    assert link.next_url == "https://example.com/query?cursor=1600000000.5_1"
    assert link.first_url == "https://example.com/query"

    header = (
        '<https://example.com/query?limit=10>; rel="first", '
        '<https://example.com/query?limit=10&cursor=15_2>; rel="next", '
        '<https://example.com/query?limit=10&cursor=p5_1>; rel="prev"'
    )
    link = PaginationLinkData.from_header(header)
    assert link.prev_url == "https://example.com/query?limit=10&cursor=p5_1"
    assert link.next_url == "https://example.com/query?limit=10&cursor=15_2"
    assert link.first_url == "https://example.com/query?limit=10"

    header = (
        '<https://example.com/query>; rel="first", '
        '<https://example.com/query?cursor=p1510000000_2>; rel="previous"'
    )
    link = PaginationLinkData.from_header(header)
    assert link.prev_url == "https://example.com/query?cursor=p1510000000_2"
    assert not link.next_url
    assert link.first_url == "https://example.com/query"

    header = '<https://example.com/query?foo=b>; rel="first"'
    link = PaginationLinkData.from_header(header)
    assert not link.prev_url
    assert not link.next_url
    assert link.first_url == "https://example.com/query?foo=b"

    link = PaginationLinkData.from_header("")
    assert not link.prev_url
    assert not link.next_url
    assert not link.first_url



"""Constants for the UWS work queue."""

from __future__ import annotations

from datetime import timedelta

__all__ = [
    "JOB_RESULT_TIMEOUT",
    "JOB_STOP_TIMEOUT",
    "UWS_DATABASE_TIMEOUT",
    "WOBBLY_REQUEST_TIMEOUT",
]

JOB_RESULT_TIMEOUT = timedelta(seconds=5)
"""How long to poll arq for job results before giving up."""

JOB_STOP_TIMEOUT = timedelta(seconds=30)
"""How long to wait for a job to stop before giving up."""

UWS_DATABASE_TIMEOUT = timedelta(seconds=30)
"""Timeout on workers that update the UWS database.

This should match the default Kubernetes grace period for a pod to shut down.
"""

WOBBLY_REQUEST_TIMEOUT = 20
"""Timeout in seconds for Wobbly HTTP requests."""



"""Manage remote Confluent schema registry schemas via Pydantic models.

Taken from Jonathan Sick's work in Kafkit_, with all
of the hard parts replaced with python-schema-registry-client_.

.. _Kafkit: https://kafkit.lsst.io/
.. _python-schema-registry-client: https://marcosschroh.github.io/python-schema-registry-client/
"""

from __future__ import annotations

import inspect
import re
from dataclasses import dataclass
from enum import StrEnum
from typing import Any

import structlog
from structlog.stdlib import BoundLogger

try:
    from dataclasses_avroschema.pydantic import AvroBaseModel
    from schema_registry.client import AsyncSchemaRegistryClient
    from schema_registry.client.errors import ClientError
    from schema_registry.serializers.message_serializer import (
        AsyncAvroMessageSerializer,
    )
except ImportError as e:
    raise ImportError(
        "The safir.kafka module requires the kafka extra. "
        "Install it with `pip install safir[kafka]`."
    ) from e

from ._exceptions import (
    IncompatibleSchemaError,
    InvalidAvroNameError,
    UnknownSchemaError,
)

__all__ = [
    "Compatibility",
    "PydanticSchemaManager",
    "SchemaInfo",
]


class Compatibility(StrEnum):
    """Schema registry compatibility types."""

    BACKWARD = "BACKWARD"
    BACKWARD_TRANSITIVE = "BACKWARD_TRANSITIVE"
    FORWARD = "FORWARD"
    FORWARD_TRANSITIVE = "FORWARD_TRANSITIVE"
    FULL = "FULL"
    FULL_TRANSITIVE = "FULL_TRANSITIVE"
    NONE = "NONE"


@dataclass
class SchemaInfo:
    """Schema and registry metadata."""

    schema: dict[str, Any]
    """The Avro schema as a Python dict."""

    schema_id: int
    """The ID of the schema in the schema registry."""

    subject: str
    """The subject that the schema is registered under in the registry."""


class PydanticSchemaManager:
    """A manager for schemas that are represented as Pydantic models in Python,
    and translated into Avro for the Confluent Schema Registry.

    It can be used to:

    * Register new schemas from Pydantic models
    * Register new versions of existing schemas from changed Pydantic models,
      after checking compatibility with the latest version in the registry

    Parameters
    ----------
    registry
        A schema registry client, or connection settings for creating one
    suffix
        A suffix that is added to the schema name (and thus subject name), for
        example ``_dev1``.

        The suffix creates alternate subjects in the Schema Registry so
        schemas registered during testing and staging don't affect the
        compatibility continuity of a production subject.

        For production, it's best to not set a suffix.
    logger
        Logger to use for internal logging. If not given, the
        ``safir.kafka.manager`` logger will be used.
    """

    def __init__(
        self,
        registry: AsyncSchemaRegistryClient,
        suffix: str = "",
        logger: BoundLogger | None = None,
    ) -> None:
        self._registry = registry
        self._serializer = AsyncAvroMessageSerializer(self._registry)
        self._suffix = suffix

        self._logger = logger or structlog.get_logger("safir.kafka.manager")

        # A mapping of subjects to registered schema ids.
        self._subjects_to_ids: dict[str, int] = {}

    async def register_model(
        self,
        model: type[AvroBaseModel],
        compatibility: Compatibility | None = None,
    ) -> SchemaInfo:
        """Register the model with the registry.

        Parameters
        ----------
        model
            The model to register.
        compatibility
            The registry `compatibility type <https://docs.confluent.io/platform/current/schema-registry/fundamentals/schema-evolution.html#compatibility-types>`__
            for this model. If not provided, it will default to the default
            type configured in the registry.

        Raises
        ------
        IncompatibleSchemaError
            If the schema is incompatible with the latest version of schema in
            the registry, according to the subject's compatibility type.
        """
        subject = self._get_subject(model)
        if compatibility:
            await self._registry.update_compatibility(
                subject=subject,
                level=compatibility,
            )

        # If a schema has been successfully registered before, and thus has a
        # version in the registry, trying to create a new version of the schema
        # in the registry will always succeed and return the schema id of the
        # previously registered version, regardless of it's compatibility
        # with the latest version. This can happen when we're trying to "roll
        # back" a schema.
        #
        # In these cases, we want to raise an error if the candidate schema is
        # incompatible. The only way to accomplish this is to explicitly check
        # compatibility with the latest version in the registry.
        #
        # NOTE: This also doesn't update the 'latest' version to the candidate
        # schema after a register call of the candidate schema. "latest"
        # remains at the last registered non-repeat version.
        schema = model.avro_schema_to_python()
        result = await self._registry.test_compatibility(
            subject=subject, schema=schema, verbose=True
        )
        # This is here just to satisfy the type checker. The test_compatibility
        # method returns a bool if verbose == False, and a dict if verbose ==
        # True.
        if not isinstance(result, dict):
            raise TypeError(
                f"test_compatibility returned {type(result).__name__}, not"
                " dict"
            )
        if "is_compatible" not in result:
            raise ValueError(
                "test_compatibility returned a dict without the"
                "'is_compatible' key: {result}"
            )

        if result["is_compatible"] is False:
            raise IncompatibleSchemaError(result["messages"])

        try:
            schema_id = await self._registry.register(subject, schema)
        except ClientError as e:
            self._logger.exception(
                "schema registry ClientError",
                msg=e.message,
                http_code=e.http_code,
                server_traceback=e.server_traceback,
            )
            raise
        self._subjects_to_ids[subject] = schema_id

        return SchemaInfo(schema=schema, schema_id=schema_id, subject=subject)

    async def serialize(self, data: AvroBaseModel) -> bytes:
        """Serialize the data.

        The model's schema must have been registered before calling this
        method.

        Parameters
        ----------
        data
            The data to serialize.

        Returns
        -------
        bytes
            The serialized data in the Confluent Wire Format.
        """
        subject = self._get_subject(data)
        try:
            schema_id = self._subjects_to_ids[subject]
        except KeyError:
            raise UnknownSchemaError(data, subject) from None
        return await self._serializer.encode_record_with_schema_id(
            schema_id, data.model_dump()
        )

    def _get_subject(self, model: AvroBaseModel | type[AvroBaseModel]) -> str:
        """Get a unique subject for this model based on metadata and config."""
        # Mypy can't detect the Meta class on the model, so we have to ignore
        # those lines.

        klass = model if inspect.isclass(model) else model.__class__

        try:
            name = str(klass.Meta.schema_name)  # type: ignore [union-attr]
        except AttributeError:
            name = klass.__name__

        try:
            namespace = str(klass.Meta.namespace)  # type: ignore [union-attr]
        except AttributeError:
            namespace = None

        if namespace:
            name = f"{namespace}.{name}"

        if self._suffix:
            name += self._suffix

        self._validate_avro_name(name)
        return name

    def _validate_avro_name(self, name: str) -> None:
        """Avro record and namespace names are constrained in their format.

        https://avro.apache.org/docs/1.11.1/specification/#names
        """
        pattern = r"^[A-Za-z_][A-Za-z0-9_]*$"

        bad = [part for part in name.split(".") if not re.match(pattern, part)]
        if bad:
            raise InvalidAvroNameError(
                f"{name} has invalid Avro name parts: {bad}. Each part must"
                f"match: {pattern}"
            )



"""Helpers working with metrics events in unit tests."""

from abc import ABC, abstractmethod
from pprint import pformat
from typing import Any, override
from unittest.mock import ANY as MOCK_ANY

from ._models import EventPayload

__all__ = [
    "ANY",
    "NOT_NONE",
    "BaseAssertionError",
    "NotPublishedConsecutivelyError",
    "NotPublishedError",
    "PublishedCountError",
    "PublishedList",
    "PublishedTooFewError",
]

ANY = MOCK_ANY
"""An object that compares equal to anything, reexported from unittest.mock."""

type ModelDumpList = list[dict[str, Any]]
"""Type alias for a list of dumped Pydantic models."""


class _NotNone:
    """A helper object that compares equal to everything except None."""

    def __eq__(self, other: object) -> bool:
        return other is not None

    def __repr__(self) -> str:
        return "<NOT NONE>"

    __hash__ = object.__hash__


NOT_NONE = _NotNone()
"""An object to indicate that a value can be anything except None."""


class BaseAssertionError[P: EventPayload](ABC, AssertionError):
    """Base assertion error with common attributes and messaging."""

    def __init__(
        self,
        expected: ModelDumpList,
        actual: ModelDumpList,
        actual_models: list[P],
    ) -> None:
        self.expected = expected
        self.actual = actual
        self.actual_models = actual_models
        message = (
            f"{self.errormsg()}\n"
            f"Expected:\n"
            f"{pformat(self.expected)}\n"
            f"Actual:\n"
            f"{pformat(self.actual)}\n"
            f"Actual models: {pformat(self.actual_models)}"
        )
        super().__init__(message)

    @abstractmethod
    def errormsg(self) -> str:
        """Return a string to be added to the exception message."""


class PublishedTooFewError[P: EventPayload](BaseAssertionError[P]):
    """Expected more events than have actually been published."""

    @override
    def errormsg(self) -> str:
        return "Expected more events than have actually been published"


class PublishedCountError[P: EventPayload](BaseAssertionError[P]):
    """Expected has a different number of items than were published."""

    @override
    def errormsg(self) -> str:
        return "Expected has a different number of items than were published"


class NotPublishedConsecutivelyError[P: EventPayload](BaseAssertionError[P]):
    """Expected events were not published consecutively."""

    @override
    def errormsg(self) -> str:
        return "Expected events were not published consecutively"


class NotPublishedError[P: EventPayload](BaseAssertionError[P]):
    """Some expected items were not published."""

    def __init__(
        self,
        *,
        expected: ModelDumpList,
        actual: ModelDumpList,
        actual_models: list[P],
        not_found: ModelDumpList,
    ) -> None:
        self.not_found = not_found
        super().__init__(expected, actual, actual_models)

    @override
    def errormsg(self) -> str:
        return (
            f"Some expected items not published\n"
            f"Not published:"
            f"{pformat(self.not_found)}"
        )


class PublishedList[P: EventPayload](list[P]):
    """A list of event payload models with assertion helpers.

    All assertion helpers take lists of dicts as expected items and use the
    ``model_dump(mode="json")`` serialization of the models for comparison.
    """

    def assert_published(
        self,
        expected: ModelDumpList,
        *,
        any_order: bool = False,
    ) -> None:
        """Assert that all of the expected payloads were published.


        Parameters
        ----------
        expected
            A list of expected event payload dicts, to be compared with the
            ``model_dump(format="json")`` serialization of the actual published
            payloads.
        any_order
            If true, then the expected payload list must be an ordered subset
            of the actual published payloads. This is like the ``any_order``
            parameter in `unittest.mock.Mock` methods.

        Raises
        ------
        PublishedTooFewError
            Expected more events than have actually been published
        NotPublishedError
            Some expected items were not published
        NotPublishedConsecutivelyError
            Expected events were not published consecutively
        """
        if len(expected) > len(self):
            actual = [model.model_dump(mode="json") for model in self]
            raise PublishedTooFewError(expected, actual, self)

        if any_order:
            self._check_unordered(expected)
        else:
            # Try unordered first to differentiate between the cases where some
            # expected events were not published at all, vs all of the expected
            # events were published, just not in the expected order.
            self._check_unordered(expected)
            self._check_ordered(expected)

    def assert_published_all(
        self,
        expected: ModelDumpList,
        *,
        any_order: bool = False,
    ) -> None:
        """Assert that all of the expected payloads, and only the expected
        payloads, were published.

        Parameters
        ----------
        expected
            A list of expected event payload dicts, to be compared with the
            ``model_dump(format="json")`` serialization of the actual published
            payloads.
        any_order
            If true, then the expected payloads must be in the same order as
            the actual published payloads. This is like the ``any_order``
            parameter in `unittest.mock.Mock` methods.

        Raises
        ------
        NotPublishedError
            Some expected items were not published
        NotPublishedConsecutivelyError
            Expected events were not published consecutively
        PublishedCountError
            A different number of events were published than were expected
        """
        if len(expected) != len(self):
            actual = [model.model_dump(mode="json") for model in self]
            raise PublishedCountError(expected, actual, self)
        self.assert_published(expected, any_order=any_order)

    def _check_unordered(
        self,
        expected: ModelDumpList,
    ) -> None:
        """Assert that each expected event was published.

        Each published event is matched only once. Taken from the
        ``unittest.mock`` ``assert_has_calls`` method.
        """
        actual = [model.model_dump(mode="json") for model in self]
        not_found = []
        tmp_actual = actual.copy()
        for item in expected:
            try:
                tmp_actual.remove(item)
            except ValueError:
                not_found.append(item)
        if not_found:
            raise NotPublishedError(
                expected=expected,
                actual=actual,
                actual_models=self,
                not_found=not_found,
            )

    def _check_ordered(self, expected: ModelDumpList) -> None:
        """Assert that the list of expected events is an ordered subset of the
        list of published events.

        Taken from the ``unittest.mock`` ``assert_has_calls`` method.
        """
        len_expected = len(expected)
        len_self = len(self)
        actual = [model.model_dump(mode="json") for model in self]
        for i in range(len_self - len_expected + 1):
            sub_list = actual[i : i + len_expected]
            if sub_list == expected:
                return
        raise NotPublishedConsecutivelyError(expected, actual, self)



"""Sentry helpers."""

from ._config import SentryConfig, initialize_sentry, should_enable_sentry
from ._helpers import (
    before_send_handler,
    duration,
    fingerprint_env_handler,
    report_exception,
    sentry_exception_handler,
)

__all__ = [
    "SentryConfig",
    "before_send_handler",
    "duration",
    "fingerprint_env_handler",
    "initialize_sentry",
    "report_exception",
    "sentry_exception_handler",
    "should_enable_sentry",
]



"""Tests for `safir.models`."""

from __future__ import annotations

import json

from safir.models import ErrorModel


def test_error_model() -> None:
    """Nothing much to test, but make sure the code can be imported."""
    error = {
        "detail": [
            {
                "loc": ["path", "foo"],
                "msg": "Invalid foo",
                "type": "invalid_foo",
            }
        ]
    }
    model = ErrorModel.model_validate_json(json.dumps(error))
    assert model.model_dump() == error



"""Date and time manipulation utility functions."""

from ._current import current_datetime
from ._format import format_datetime_for_logging, isodatetime
from ._parse import parse_isodatetime, parse_timedelta

__all__ = [
    "current_datetime",
    "format_datetime_for_logging",
    "isodatetime",
    "parse_isodatetime",
    "parse_timedelta",
]



"""Construction of UWS backend workers."""

import asyncio
import os
import signal
import uuid
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from datetime import UTC, datetime, timedelta
from enum import Enum
from traceback import format_exception
from typing import Any, ClassVar
from urllib.parse import urlsplit

from arq import func
from arq.connections import RedisSettings
from pydantic import BaseModel
from structlog.stdlib import BoundLogger

from . import ArqMode, ArqQueue, MockArqQueue, RedisArqQueue, WorkerSettings

UWS_QUEUE_NAME = "uws:queue"
"""Name of the arq queue for internal UWS messages."""

__all__ = [
    "UWS_QUEUE_NAME",
    "WorkerConfig",
    "WorkerError",
    "WorkerErrorType",
    "WorkerFatalError",
    "WorkerJobInfo",
    "WorkerResult",
    "WorkerTimeoutError",
    "WorkerTransientError",
    "WorkerUsageError",
    "build_worker",
]


@dataclass
class WorkerConfig[T: BaseModel]:
    """Minimal configuration needed for building a UWS backend worker."""

    arq_mode: ArqMode
    """What mode to use for the arq queue."""

    arq_queue_url: str
    """URL of the Redis arq queue."""

    arq_queue_password: str | None
    """Password of the Redis arq queue."""

    grace_period: timedelta
    """How long to wait for workers to shut down before cancelling them.

    This should be set to somewhat less than the Kubernetes grace period for
    terminating the pod (about five seconds less, for example).
    """

    parameters_class: type[T]
    """Class of the parameters to pass to the backend worker."""

    timeout: timedelta
    """Maximum execution time.

    Jobs that run longer than this length of time will be automatically
    aborted.
    """

    @property
    def arq_redis_settings(self) -> RedisSettings:
        """Redis settings for arq."""
        database = 0
        url = urlsplit(self.arq_queue_url)
        if url.path:
            database = int(url.path.lstrip("/"))
        return RedisSettings(
            host=url.hostname or "localhost",
            port=url.port or 6379,
            database=database,
            password=self.arq_queue_password,
        )


@dataclass
class WorkerJobInfo:
    """Metadata about the job that may be useful to the backend."""

    job_id: str
    """UWS job identifier (not the same as the arq job ID)."""

    user: str
    """Username of the user who submitted the job."""

    token: str
    """Delegated Gafaelfawr token to act on behalf of the user."""

    timeout: timedelta
    """Maximum execution time for the job."""

    run_id: str | None = None
    """User-supplied run ID, if any."""


class WorkerResult(BaseModel):
    """A single result from the job."""

    result_id: str
    """Identifier for the result."""

    url: str
    """URL for the result, which must point to a GCS bucket."""

    size: int | None = None
    """Size of the result in bytes."""

    mime_type: str | None = None
    """MIME type of the result."""


class WorkerErrorType(Enum):
    """Types of errors that may be reported by a worker."""

    FATAL = "fatal"
    TRANSIENT = "transient"
    USAGE = "usage"


class WorkerError(Exception):
    """An error occurred during background task processing.

    Attributes
    ----------
    detail
        Additional error detail, not including the traceback if any.
    user
        User whose action triggered this exception, for Slack reporting.

    Parameters
    ----------
    message
        Human-readable error message.
    detail
        Additional details about the error.
    add_traceback
        Whether to add a traceback of the underlying cause to the error
        details.
    """

    error_type: ClassVar[WorkerErrorType] = WorkerErrorType.FATAL
    """Type of error this exception represents."""

    def __init__(
        self,
        message: str,
        detail: str | None = None,
        *,
        add_traceback: bool = False,
    ) -> None:
        super().__init__(message)
        self.detail = detail
        self._cause_type: str | None = None
        self._traceback: str | None = None
        self._add_traceback = add_traceback

    def __reduce__(self) -> str | tuple:
        # Ensure the cause information is serialized before pickling.
        self._cause_type = self._serialize_cause_type()
        self._traceback = self._serialize_traceback()
        return super().__reduce__()

    @property
    def cause_type(self) -> str | None:
        """Type of the exception that triggered this error, if known."""
        if not self._cause_type:
            self._cause_type = self._serialize_cause_type()
        return self._cause_type

    @property
    def traceback(self) -> str | None:
        """Traceback of the underlying exception, if desired."""
        if not self._traceback:
            self._traceback = self._serialize_traceback()
        return self._traceback

    def _serialize_cause_type(self) -> str | None:
        """Serialize the type of exception from ``__cause__``."""
        if not self.__cause__:
            return None
        return type(self.__cause__).__qualname__

    def _serialize_traceback(self) -> str | None:
        """Serialize the traceback from ``__cause__``."""
        if not self._add_traceback or not self.__cause__:
            return None
        return "".join(format_exception(self.__cause__))


class WorkerFatalError(WorkerError):
    """Fatal error occurred during worker processing.

    The parameters or other job information was invalid and this job will
    never succeed.
    """


class WorkerTransientError(WorkerError):
    """Transient error occurred during worker processing.

    The job may be retried with the same parameters and may succeed.
    """

    error_type = WorkerErrorType.TRANSIENT


class WorkerTimeoutError(WorkerTransientError):
    """Transient error occurred during worker processing.

    The job may be retried with the same parameters and may succeed.
    """

    def __init__(self, elapsed: timedelta, timeout: timedelta) -> None:
        msg = (
            f"Job timed out after {elapsed.total_seconds()}s"
            f" (timeout: {timeout.total_seconds()}s)"
        )
        super().__init__(msg)


class WorkerUsageError(WorkerError):
    """Parameters sent by the user were invalid.

    The parameters or other job information was invalid and this job will
    never succeed. This is the same as `WorkerFatalError` except that it
    represents a user error and will not be reported to Slack as a service
    problem.
    """

    error_type = WorkerErrorType.USAGE


def _restart_pool(pool: ProcessPoolExecutor) -> ProcessPoolExecutor:
    """Restart the pool after timeout or job cancellation.

    This is a horrible, fragile hack, but it appears to be the only way to
    enforce a timeout currently in Python since there is no way to abort a
    job already in progress. Find the processes underlying the pool, kill
    them, and then shut down and recreate the pool.
    """
    for pid in pool._processes:  # noqa: SLF001
        os.kill(pid, signal.SIGINT)
    pool.shutdown(wait=True)
    return ProcessPoolExecutor(1)


def build_worker[T: BaseModel](
    worker: Callable[[T, WorkerJobInfo, BoundLogger], list[WorkerResult]],
    config: WorkerConfig[T],
    logger: BoundLogger,
) -> WorkerSettings:
    """Construct an arq worker for the provided backend function.

    Builds an arq worker configuration that wraps the provided sync function
    and executes it on messages to the default arq queue. Messages to the UWS
    queue will be sent on job start and after job completion so that the UWS
    database can be updated.

    Unfortunately, the built-in arq ``on_job_start`` and ``after_job_end``
    hooks can't be used because they don't receive any arguments to the job
    and we need to tell the UWS handlers the job ID to act on. This means that
    we'll send the UWS queue message before the results are recorded in Redis,
    so the UWS handler has to deal with that.

    Parameters
    ----------
    worker
        Synchronous function that does the actual work. This function will be
        run in a thread pool of size one.
    config
        UWS worker configuration.
    logger
        Logger to use for messages.
    """

    async def startup(ctx: dict[Any, Any]) -> None:
        nonlocal logger
        logger = logger.bind(worker_instance=uuid.uuid4().hex)

        # The queue to which to send UWS notification messages.
        if config.arq_mode == ArqMode.production:
            settings = config.arq_redis_settings
            arq: ArqQueue = await RedisArqQueue.initialize(
                settings, default_queue_name=UWS_QUEUE_NAME
            )
        else:
            arq = MockArqQueue(default_queue_name=UWS_QUEUE_NAME)

        ctx["arq"] = arq
        ctx["logger"] = logger
        ctx["pool"] = ProcessPoolExecutor(1)

        logger.info("Worker startup complete")

    async def shutdown(ctx: dict[Any, Any]) -> None:
        logger: BoundLogger = ctx["logger"]
        pool: ProcessPoolExecutor = ctx["pool"]

        pool.shutdown(wait=True, cancel_futures=True)

        logger.info("Worker shutdown complete")

    async def run(
        ctx: dict[Any, Any], params_raw: dict[str, Any], info: WorkerJobInfo
    ) -> list[WorkerResult]:
        arq: ArqQueue = ctx["arq"]
        logger: BoundLogger = ctx["logger"]
        pool: ProcessPoolExecutor = ctx["pool"]

        params = config.parameters_class.model_validate(params_raw)
        logger = logger.bind(
            task=worker.__qualname__,
            job_id=info.job_id,
            user=info.user,
            params=params.model_dump(mode="json"),
        )
        if info.run_id:
            logger = logger.bind(run_id=info.run_id)

        start = datetime.now(tz=UTC)
        await arq.enqueue("uws_job_started", info.token, info.job_id, start)
        loop = asyncio.get_running_loop()
        try:
            async with asyncio.timeout(info.timeout.total_seconds()):
                return await loop.run_in_executor(
                    pool, worker, params, info, logger
                )
        except asyncio.CancelledError:
            ctx["pool"] = _restart_pool(pool)
            raise
        except TimeoutError:
            elapsed = datetime.now(tz=UTC) - start
            ctx["pool"] = _restart_pool(pool)
            raise WorkerTimeoutError(elapsed, info.timeout) from None
        finally:
            await arq.enqueue("uws_job_completed", info.token, info.job_id)

    # Since the worker is running sync jobs, run one job per pod since they
    # will be serialized anyway and no parallelism is possible. This also
    # allows us to easily restart the job pool on timeout or job abort. If
    # async worker support is added, consider making this configurable.
    return WorkerSettings(
        functions=[func(run, name=worker.__qualname__)],
        redis_settings=config.arq_redis_settings,
        job_completion_wait=config.grace_period,
        job_timeout=config.timeout,
        max_jobs=1,
        allow_abort_jobs=True,
        on_startup=startup,
        on_shutdown=shutdown,
    )



"""Test IVOA middleware."""

from __future__ import annotations

from collections.abc import AsyncGenerator
from typing import Annotated

import pytest
import pytest_asyncio
from fastapi import FastAPI, Query, Request
from httpx import ASGITransport, AsyncClient

from safir.middleware.ivoa import (
    CaseInsensitiveFormMiddleware,
    CaseInsensitiveQueryMiddleware,
)


def build_app() -> FastAPI:
    """Construct a test FastAPI app with the middleware registered."""
    app = FastAPI()
    app.add_middleware(CaseInsensitiveQueryMiddleware)
    app.add_middleware(CaseInsensitiveFormMiddleware)

    @app.get("/")
    async def handler(param: str) -> dict[str, str]:
        return {"param": param}

    @app.get("/simple")
    async def simple_handler() -> dict[str, str]:
        return {"foo": "bar"}

    @app.get("/list")
    async def list_handler(
        param: Annotated[list[str], Query()],
    ) -> dict[str, list[str]]:
        return {"param": param}

    @app.post("/form-list")
    async def form_handler(request: Request) -> dict[str, list[str]]:
        form = await request.form()
        return {
            "param": [str(v) for v in form.getlist("param")],
            "received_keys": list(form.keys()),
        }

    return app


@pytest_asyncio.fixture
async def client() -> AsyncGenerator[AsyncClient, None]:
    """Test client fixture with the IVOA middleware configured."""
    app = build_app()
    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        yield client


@pytest.mark.asyncio
async def test_single_query_param_case_insensitive(
    client: AsyncClient,
) -> None:
    """Test that single query parameters are handled case-insensitively."""
    # Test normal case
    r = await client.get("/", params={"param": "foo"})
    assert r.status_code == 200
    assert r.json() == {"param": "foo"}

    # Test uppercase parameter
    r = await client.get("/", params={"PARAM": "foo"})
    assert r.status_code == 200
    assert r.json() == {"param": "foo"}

    # Test mixed case parameter
    r = await client.get("/", params={"pARam": "foo"})
    assert r.status_code == 200
    assert r.json() == {"param": "foo"}


@pytest.mark.asyncio
async def test_query_param_error_handling(client: AsyncClient) -> None:
    """Test error handling for invalid query parameters."""
    r = await client.get("/", params={"paramX": "foo"})
    assert r.status_code == 422


@pytest.mark.asyncio
async def test_simple_endpoint(client: AsyncClient) -> None:
    """Test endpoint with no parameters."""
    r = await client.get("/simple")
    assert r.status_code == 200
    assert r.json() == {"foo": "bar"}


@pytest.mark.asyncio
async def test_list_query_params_case_insensitive(client: AsyncClient) -> None:
    """Test that list query parameters are handled case-insensitively."""
    r = await client.get(
        "/list",
        params=[("param", "foo"), ("PARAM", "BAR"), ("parAM", "baZ")],
    )
    assert r.status_code == 200
    assert r.json() == {"param": ["foo", "BAR", "baZ"]}


@pytest.mark.asyncio
async def test_form_data_case_insensitive(client: AsyncClient) -> None:
    """Test that form data parameters are handled case-insensitively."""
    form_data = {"param": "foo", "PARAM": "BAR", "parAM": "baZ"}
    r = await client.post("/form-list", data=form_data)
    assert r.status_code == 200
    response_data = r.json()
    assert response_data["param"] == ["foo", "BAR", "baZ"]
    assert all(key == "param" for key in response_data["received_keys"])


@pytest.mark.asyncio
async def test_empty_form_data(client: AsyncClient) -> None:
    """Test that the endpoint handles empty form data gracefully."""
    r = await client.post("/form-list", data={})
    assert r.status_code == 200
    response_data = r.json()
    assert response_data["param"] == []
    assert response_data["received_keys"] == []



"""Test Slack client."""

from __future__ import annotations

from unittest.mock import ANY

import pytest
import structlog
from fastapi import APIRouter, FastAPI
from httpx import ASGITransport, AsyncClient
from pydantic import SecretStr

from safir.datetime import current_datetime, format_datetime_for_logging
from safir.slack.blockkit import SlackException, SlackMessage
from safir.slack.webhook import (
    SlackIgnoredException,
    SlackRouteErrorHandler,
    SlackWebhookClient,
)
from safir.testing.slack import MockSlackWebhook


@pytest.mark.asyncio
async def test_post(mock_slack: MockSlackWebhook) -> None:
    message = SlackMessage(message="Some random message")
    logger = structlog.get_logger(__file__)
    client = SlackWebhookClient(mock_slack.url, "App", logger)
    await client.post(message)
    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Some random message",
                        "verbatim": True,
                    },
                }
            ]
        }
    ]


@pytest.mark.asyncio
async def test_post_exception(mock_slack: MockSlackWebhook) -> None:
    logger = structlog.get_logger(__file__)
    client = SlackWebhookClient(SecretStr(mock_slack.url), "App", logger)

    exc = SlackException("Some exception message")
    await client.post_exception(exc)
    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App: Some exception message",
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nSlackException",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": ANY,
                            "verbatim": True,
                        },
                    ],
                },
                {"type": "divider"},
            ]
        }
    ]
    mock_slack.messages = []

    class TestException(SlackException):
        pass

    now = current_datetime()
    now_formatted = format_datetime_for_logging(now)
    exc = TestException("Blah blah blah", "username", failed_at=now)
    await client.post_exception(exc)
    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App: Blah blah blah",
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nTestException",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": f"*Failed at*\n{now_formatted}",
                            "verbatim": True,
                        },
                        {
                            "type": "mrkdwn",
                            "text": "*User*\nusername",
                            "verbatim": True,
                        },
                    ],
                },
                {"type": "divider"},
            ]
        }
    ]


@pytest.mark.asyncio
async def test_route_handler(mock_slack: MockSlackWebhook) -> None:
    """Test Slack alerts for uncaught exceptions."""
    router = APIRouter(route_class=SlackRouteErrorHandler)

    class SomeAppError(SlackIgnoredException):
        pass

    class OtherAppError(SlackException):
        pass

    @router.get("/exception")
    async def get_exception() -> None:
        raise ValueError("Test exception")

    @router.get("/ignored")
    async def get_ignored() -> None:
        raise SomeAppError("Test exception")

    @router.get("/known")
    async def get_known() -> None:
        raise OtherAppError("Slack exception")

    app = FastAPI()
    app.include_router(router)

    # We need a custom HTTPX configuration to disable raising server
    # exceptions so that we can inspect the resulting error handling.
    transport = ASGITransport(app=app, raise_app_exceptions=False)

    # Run the test.
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get("/exception")
        assert r.status_code == 500

        # Slack alerting has not been configured yet, so nothing should be
        # posted to Slack.
        assert mock_slack.messages == []

        # Configure Slack alerting and raise an ignored exception. There
        # should still be nothing posted to Slack.
        logger = structlog.get_logger(__file__)
        SlackRouteErrorHandler.initialize(mock_slack.url, "App", logger)
        r = await client.get("/ignored")
        assert r.status_code == 500
        assert mock_slack.messages == []

        # But now raising another exception should result in a Slack message.
        r = await client.get("/exception")
        assert r.status_code == 500

        # And raising an exception that inherits from SlackException should
        # result in nicer formatting.
        r = await client.get("/known")
        assert r.status_code == 500

    # Check that the Slack alert was what we expected.
    assert mock_slack.messages == [
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": "Error in App",
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nValueError",
                            "verbatim": True,
                        },
                        {"type": "mrkdwn", "text": ANY, "verbatim": True},
                    ],
                },
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": (
                            "*Exception*\n```\nValueError: Test exception\n```"
                        ),
                        "verbatim": True,
                    },
                },
                {"type": "divider"},
            ],
        },
        {
            "blocks": [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": ("Error in App: Slack exception"),
                        "verbatim": True,
                    },
                },
                {
                    "type": "section",
                    "fields": [
                        {
                            "type": "mrkdwn",
                            "text": "*Exception type*\nOtherAppError",
                            "verbatim": True,
                        },
                        {"type": "mrkdwn", "text": ANY, "verbatim": True},
                    ],
                },
                {"type": "divider"},
            ],
        },
    ]
    assert mock_slack.messages[0]["blocks"][1]["fields"][1]["text"].startswith(
        "*Failed at*\n"
    )



"""Test metrics configuration."""

from __future__ import annotations

import pytest
from pydantic import Field
from pydantic_settings import BaseSettings

from safir.metrics import (
    DisabledMetricsConfiguration,
    EventsConfiguration,
    KafkaEventManager,
    KafkaMetricsConfiguration,
    MetricsConfiguration,
    MockEventManager,
    MockMetricsConfiguration,
    NoopEventManager,
    metrics_configuration_factory,
)
from tests.support.kafka import KafkaClients, KafkaStack


class Config(BaseSettings):
    metrics: MetricsConfiguration = Field(
        default_factory=metrics_configuration_factory
    )


def test_disabled(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("METRICS_APPLICATION", "test")
    monkeypatch.setenv("METRICS_ENABLED", "false")

    config = Config()
    assert isinstance(config.metrics, DisabledMetricsConfiguration)
    manager = config.metrics.make_manager()
    assert isinstance(manager, NoopEventManager)


def test_disabled_extra() -> None:
    config = Config.model_validate(
        {
            "metrics": {
                "enabled": False,
                "application": "example",
                "events": {"topicPrefix": "lsst.square.metrics.events"},
                "schemaManager": {
                    "registryUrl": "https://example.com",
                    "suffix": "",
                },
            }
        }
    )
    assert isinstance(config.metrics, DisabledMetricsConfiguration)


def test_mock(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("METRICS_APPLICATION", "test")
    monkeypatch.setenv("METRICS_ENABLED", "false")
    monkeypatch.setenv("METRICS_MOCK", "true")

    config = Config()
    assert isinstance(config.metrics, MockMetricsConfiguration)
    manager = config.metrics.make_manager()
    assert isinstance(manager, MockEventManager)


@pytest.mark.asyncio
async def test_kafka(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("METRICS_APPLICATION", "test")
    monkeypatch.setenv("KAFKA_SECURITY_PROTOCOL", "PLAINTEXT")
    monkeypatch.setenv("KAFKA_BOOTSTRAP_SERVERS", "kafka-1.sasquatch:9092")
    monkeypatch.setenv("SCHEMA_MANAGER_REGISTRY_URL", "https://example.com/")

    config = Config()
    assert isinstance(config.metrics, KafkaMetricsConfiguration)
    manager = config.metrics.make_manager()
    assert isinstance(manager, KafkaEventManager)


@pytest.mark.asyncio
async def test_kafka_unmanaged(
    kafka_stack: KafkaStack, kafka_clients: KafkaClients
) -> None:
    config = KafkaMetricsConfiguration(
        application="testapp",
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=kafka_stack.kafka_connection_settings,
        schema_manager=kafka_stack.schema_manager_settings,
    )

    manager = config.make_manager(kafka_broker=kafka_clients.kafka_broker)
    assert isinstance(manager, KafkaEventManager)

    # Make sure that the manager doesn't close the broker.
    await manager.aclose()
    assert await manager._broker.ping(timeout=1)



"""Schema manager exceptions."""

from __future__ import annotations

from pydantic import SecretStr
from structlog.stdlib import BoundLogger

from ..sentry import report_exception
from ..slack.webhook import SlackWebhookClient

try:
    from dataclasses_avroschema.pydantic import AvroBaseModel
    from faststream import ExceptionMiddleware
except ImportError as e:
    raise ImportError(
        "The safir.kafka module requires the kafka extra. "
        "Install it with `pip install safir[kafka]`."
    ) from e

__all__ = [
    "FastStreamErrorHandler",
    "IncompatibleSchemaError",
    "InvalidAvroNameError",
    "InvalidMetadataError",
    "UnknownDeserializeError",
    "UnknownSchemaError",
]


class FastStreamErrorHandler:
    """Error handler for FastStream that reports errors to external services.

    Initializes and creates a `FastStream ExceptionMiddleware`_ handler that
    reports any uncaught exceptions in FastStream message handlers to any
    exception reporting services that are configured.

    * To report errors directly to Slack,
      `FastStreamErrorHandler.initialize_slack` must be called with a valid
      Slack webhook before any messages are handled.
    * To report errors to Sentry, Sentry has to be initialized before any
      messages are handled.

    .. _FastStream ExceptionMiddleware: https://faststream.ag2.ai/latest/getting-started/middlewares/exception

    Examples
    --------
    Use this class when creating a ``KafkaRouter``.

    .. code-block:: python

       from safir.kafka import FastStreamErrorHandler
       from safir.sentry import initialize_sentry

       # If a Sentry DSN is provided in an environment variable, then
       # exceptions will be reported to Sentry.
       initialize_sentry()

       error_handler = FastStreamErrorHandler()

       kafka_router = KafkaRouter(
           middlewares=[error_handler.make_middleware()],
           **kafka_params,
           logger=logger,
       )

       logger = get_logger("my-app")

       if config.slack.enabled:
           error_handler.initialize_slack("some_webhook", "my-app", logger)
    """

    def __init__(self) -> None:
        self._slack_client: SlackWebhookClient | None = None

    def initialize_slack(
        self, hook_url: str | SecretStr, application: str, logger: BoundLogger
    ) -> None:
        """Configure Slack alerting.

        Until this function is called, all Slack alerting for uncaught
        exceptions will be disabled.

        Parameters
        ----------
        hook_url
            The URL of the incoming webhook to use to publish the message.
        application
            Name of the application reporting an error.
        logger
            Logger to which to report errors sending messages to Slack.
        """
        self._slack_client = SlackWebhookClient(hook_url, application, logger)

    async def handle_error(self, exc: Exception) -> None:
        """Report an exception to Slack and/or Sentry and re-raise it.

        This will called by the FastStream framework when the middleware
        generated by `FastStreamErrorHandler.make_middleware` is added to a
        router or broker.
        """
        await report_exception(exc, slack_client=self._slack_client)
        raise exc

    def make_middleware(self) -> ExceptionMiddleware:
        """Create a FastStream Exception middleware to add to a broker."""
        return ExceptionMiddleware(handlers={Exception: self.handle_error})


class IncompatibleSchemaError(Exception):
    """A schema is incompatible with the latest version in the registry."""


class InvalidAvroNameError(Exception):
    """The decalred name or namespace for an Avro schema is not valid."""


class InvalidMetadataError(Exception):
    """The Meta inner class on a model has unexpected values in fields."""


class UnknownDeserializeError(Exception):
    """The schema registry client returns None when deserializing."""


class UnknownSchemaError(Exception):
    """A schema is not managed by the Registry, and therefore cannot be
    serialized into a native Python object.
    """

    def __init__(self, data: AvroBaseModel, subject: str) -> None:
        self.message = (
            f"Schema for model: {type(data).__name__} with subject: {subject}"
            " is not known to the manager. ``register`` must be called before"
            " you try to serialize instances of this model."
        )
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



"""Models for structlog-based logging."""

from __future__ import annotations

from enum import Enum
from typing import Any, Self

__all__ = [
    "LogLevel",
    "Profile",
]


class Profile(Enum):
    """Logging profile for the application."""

    production = "production"
    """Log messages in JSON."""

    development = "development"
    """Log messages in a format intended for human readability."""


class LogLevel(Enum):
    """Python logging level.

    Any case variation is accepted when converting a string to an enum value
    via the class constructor.
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"

    @classmethod
    def _missing_(cls, value: Any) -> Self | None:
        """Allow strings in any case to be used to create the enum."""
        if not isinstance(value, str):
            return None
        value = value.upper()
        for member in cls:
            if member.value == value:
                return member
        return None



"""Tests for the safir.testing.logging module."""

from __future__ import annotations

import json
import logging
import uuid
from datetime import UTC, datetime

import pytest
import structlog

from safir.datetime import isodatetime
from safir.logging import LogLevel, Profile, configure_logging
from safir.testing.logging import parse_log_tuples


def test_parse_log_tuples(
    caplog: pytest.LogCaptureFixture,
) -> None:
    configure_logging(
        name="myapp",
        profile=Profile.production,
        log_level=LogLevel.INFO,
        add_timestamp=True,
    )

    logger = structlog.get_logger("myapp")
    logger = logger.bind(answer=42)
    logger.info("Hello world")

    assert parse_log_tuples("myapp", caplog.record_tuples) == [
        {
            "answer": 42,
            "event": "Hello world",
            "severity": "info",
        }
    ]
    assert parse_log_tuples("otherapp", caplog.record_tuples) == []

    test_tuples = [
        (
            "myapp",
            logging.INFO,
            json.dumps(
                {
                    "event": "Some message",
                    "httpRequest": {
                        "remoteIp": "127.0.0.1",
                        "requestMethod": "GET",
                        "requestUrl": "https://example.com/",
                        "userAgent": "some-user-agent/1.0",
                    },
                    "logger": "myapp",
                    "request_id": str(uuid.uuid4()),
                    "severity": "info",
                    "timestamp": isodatetime(datetime.now(tz=UTC)),
                }
            ),
        ),
        (
            "myapp",
            logging.DEBUG,
            json.dumps(
                {
                    "event": "Debug message",
                    "logger": "myapp",
                    "severity": "debug",
                }
            ),
        ),
        (
            "otherapp",
            logging.WARNING,
            json.dumps(
                {
                    "event": "Warning message",
                    "logger": "otherapp",
                    "severity": "warning",
                }
            ),
        ),
    ]
    expected = [
        {
            "event": "Some message",
            "httpRequest": {
                "remoteIp": "127.0.0.1",
                "requestMethod": "GET",
                "requestUrl": "https://example.com/",
            },
            "severity": "info",
        },
        {
            "event": "Debug message",
            "severity": "debug",
        },
    ]
    assert parse_log_tuples("myapp", test_tuples) == expected
    filtered = parse_log_tuples("myapp", test_tuples, ignore_debug=True)
    assert filtered == expected[0:1]

    with pytest.raises(AssertionError, match="logger "):
        parse_log_tuples(
            "myapp", [("myapp", logging.INFO, json.dumps({"logger": "foo"}))]
        )
    with pytest.raises(AssertionError, match="severity "):
        parse_log_tuples(
            "myapp",
            [
                (
                    "myapp",
                    logging.INFO,
                    json.dumps({"logger": "myapp", "severity": "warning"}),
                )
            ],
        )
    with pytest.raises(AssertionError, match="endswith"):
        parse_log_tuples(
            "myapp",
            [
                (
                    "myapp",
                    logging.INFO,
                    json.dumps(
                        {
                            "event": "Message",
                            "logger": "myapp",
                            "severity": "info",
                            "timestamp": "2025-05-27T12:23:45",
                        }
                    ),
                )
            ],
        )
    with pytest.raises(AssertionError, match="within 1 hour"):
        parse_log_tuples(
            "myapp",
            [
                (
                    "myapp",
                    logging.INFO,
                    json.dumps(
                        {
                            "event": "Message",
                            "logger": "myapp",
                            "severity": "info",
                            "timestamp": "2025-05-27T10:23:45Z",
                        }
                    ),
                )
            ],
        )



"""nox build configuration for Safir."""

from __future__ import annotations

import shutil
import sys
from pathlib import Path

import nox
from nox.command import CommandFailed
from nox_uv import session

# Default sessions
nox.options.sessions = ["lint", "typing", "test", "docs", "docs-linkcheck"]

# Other nox defaults
nox.options.default_venv_backend = "uv"
nox.options.reuse_existing_virtualenvs = True

# Extras to include for most actions. This should be all of the extras from
# the top-level pyproject.toml.
_EXTRAS = [
    "arq",
    "db",
    "gcs",
    "kafka",
    "kubernetes",
    "redis",
    "testcontainers",
    "uws",
]


@session(uv_groups=["dev", "docs"], uv_extras=_EXTRAS)
def docs(session: nox.Session) -> None:
    """Build the documentation."""
    doctree_dir = (session.cache_dir / "doctrees").absolute()
    with session.chdir("docs"):
        session.run(
            "sphinx-build",
            "-W",
            "--keep-going",
            "-n",
            "-T",
            "-b",
            "html",
            "-d",
            str(doctree_dir),
            ".",
            "./_build/html",
        )


@session(name="docs-clean", uv_groups=["dev", "docs"], uv_extras=_EXTRAS)
def docs_clean(session: nox.Session) -> None:
    """Build the documentation without any cache."""
    if Path("docs/_build").exists():
        shutil.rmtree("docs/_build")
    if Path("docs/api").exists():
        shutil.rmtree("docs/api")
    docs(session)


@session(name="docs-linkcheck", uv_groups=["dev", "docs"], uv_extras=_EXTRAS)
def docs_linkcheck(session: nox.Session) -> None:
    """Check links in the documentation."""
    doctree_dir = (session.cache_dir / "doctrees").absolute()
    with session.chdir("docs"):
        try:
            session.run(
                "sphinx-build",
                "-W",
                "--keep-going",
                "-n",
                "-T",
                "-b",
                "linkcheck",
                "-d",
                str(doctree_dir),
                ".",
                "./_build/linkcheck",
            )
        except CommandFailed:
            output_path = Path("_build") / "linkcheck" / "output.txt"
            if output_path.exists():
                sys.stdout.write(output_path.read_text())
            session.error("Link check reported errors")


@session(uv_only_groups=["lint"], uv_no_install_project=True)
def lint(session: nox.Session) -> None:
    """Run pre-commit hooks."""
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@session(uv_groups=["dev"], uv_extras=_EXTRAS)
def test(session: nox.Session) -> None:
    """Run the test suite."""
    session.run("pytest", *session.posargs)


@session(uv_groups=["dev", "typing"], uv_extras=_EXTRAS)
def typing(session: nox.Session) -> None:
    """Check type annotations with mypy."""
    session.run(
        "mypy",
        *session.posargs,
        "--namespace-packages",
        "--explicit-package-bases",
        "noxfile.py",
        "safir-arq/src",
        "safir-logging/src",
        "src",
        "tests",
        env={"MYPYPATH": "safir-arq/src:safir-logging/src:src"},
    )



"""Database initialization."""

from __future__ import annotations

import asyncio

try:
    from sqlalchemy.exc import OperationalError
    from sqlalchemy.ext.asyncio import AsyncEngine
    from sqlalchemy.schema import CreateSchema
    from sqlalchemy.sql.schema import MetaData
except ImportError as e:
    raise ImportError(
        "The safir.database module requires the db extra. "
        "Install it with `pip install safir[db]`."
    ) from e
from structlog.stdlib import BoundLogger

from ._alembic import unstamp_database

__all__ = [
    "DatabaseInitializationError",
    "drop_database",
    "initialize_database",
]


class DatabaseInitializationError(Exception):
    """Database initialization failed."""


async def drop_database(engine: AsyncEngine, schema: MetaData) -> None:
    """Drop all tables from the database.

    Parameters
    ----------
    engine
        Engine to use to issue the SQL commands.
    schema
        Database ORM schema.
    """
    async with engine.begin() as conn:
        await conn.run_sync(schema.drop_all)
    await unstamp_database(engine)


async def initialize_database(
    engine: AsyncEngine,
    logger: BoundLogger,
    *,
    schema: MetaData,
    reset: bool = False,
) -> None:
    """Create and initialize a new database.

    Parameters
    ----------
    engine
        Database engine to use.  Create with `create_database_engine`.
    logger
        Logger used to report problems
    schema
        Metadata for the database schema.  Generally this will be
        ``Base.metadata`` where ``Base`` is the declarative base used as the
        base class for all ORM table definitions.  The caller must ensure that
        all table definitions have been imported by Python before calling this
        function, or parts of the schema will be missing.
    reset
        If set to `True`, drop all tables and reprovision the database.
        Useful when running tests with an external database.

    Raises
    ------
    DatabaseInitializationError
        After five attempts, the database still could not be initialized.
        This is normally due to some connectivity issue to the database.
    """
    success = False
    error = None
    for _ in range(5):
        try:
            async with engine.begin() as conn:
                if schema.schema is not None:
                    await conn.execute(CreateSchema(schema.schema, True))
                if reset:
                    await conn.run_sync(schema.drop_all)
                await conn.run_sync(schema.create_all)
            success = True
        except (ConnectionRefusedError, OperationalError, OSError) as e:
            logger.info("database not ready, waiting two seconds")
            error = str(e)
            await asyncio.sleep(2)
            continue
        if success:
            logger.info("initialized database schema")
            break
    if not success:
        msg = "database schema initialization failed (database not reachable?)"
        logger.error(msg)
        await engine.dispose()
        raise DatabaseInitializationError(error)



"""A FastAPI dependency that supplies a Redis connection for arq_."""

from __future__ import annotations

from arq.connections import RedisSettings

from safir.arq import ArqMode, ArqQueue, MockArqQueue, RedisArqQueue

__all__ = ["ArqDependency", "arq_dependency"]


class ArqDependency:
    """A FastAPI dependency that maintains a Redis client for enqueuing tasks
    to the worker pool.
    """

    def __init__(self) -> None:
        self._arq_queue: ArqQueue | None = None

    async def initialize(
        self, *, mode: ArqMode, redis_settings: RedisSettings | None
    ) -> None:
        """Initialize the dependency (call during the FastAPI start-up event).

        Parameters
        ----------
        mode
            The mode to operate the queue dependency in. With
            `safir.arq.ArqMode.production`, this method initializes a
            Redis-based arq queue and the dependency creates a
            `safir.arq.RedisArqQueue` client.

            With `safir.arq.ArqMode.test`, this method instead initializes an
            in-memory mocked version of arq that you use with the
            `safir.arq.MockArqQueue` client.
        redis_settings
            The arq Redis settings, required when the ``mode`` is
            `safir.arq.ArqMode.production`. See arq's
            `~arq.connections.RedisSettings` documentation for details on
            this object.

        Examples
        --------
        .. code-block:: python

           from collections.abc import AsyncGenerator
           from contextlib import asynccontextmanager

           from fastapi import Depends, FastAPI
           from safir.arq import ArqMode, ArqQueue
           from safir.dependencies.arq import arq_dependency


           @asynccontextmanager
           async def lifespan(app: FastAPI) -> AsyncGenerator[None]:
               await arq_dependency.initialize(mode=ArqMode.test)
               yield


           app = FastAPI()


           @app.post("/")
           async def post_job(
               arq_queue: ArqQueue = Depends(arq_dependency),
           ) -> dict[str, Any]:
               job = await arq_queue.enqueue("test_task", "hello", an_int=42)
               return {"job_id": job.id}
        """
        if mode == ArqMode.production:
            if not redis_settings:
                raise RuntimeError(
                    "The redis_settings argument must be set for arq in "
                    "production."
                )
            self._arq_queue = await RedisArqQueue.initialize(redis_settings)
        else:
            self._arq_queue = MockArqQueue()

    async def __call__(self) -> ArqQueue:
        """Get the arq queue.

        This method is called for your by ``fastapi.Depends``.
        """
        if self._arq_queue is None:
            raise RuntimeError("ArqDependency is not initialized")
        return self._arq_queue


arq_dependency = ArqDependency()
"""Singleton instance of `ArqDependency` that serves as a FastAPI
dependency.
"""



"""Assert helpers for testing services using the Safir UWS support."""

from __future__ import annotations

try:
    from vo_models.uws import JobSummary
except ImportError as e:
    raise ImportError(
        "The safir.testing.uws module requires the uws extra. "
        "Install it with `pip install safir[uws]`."
    ) from e

__all__ = ["assert_job_summary_equal"]


def assert_job_summary_equal(
    job_summary_type: type[JobSummary], seen: str, expected: str
) -> None:
    """Assert that two job XML documents are equal.

    The comparison is done by converting both to a
    `~vo_models.uws.models.JobSummary` object qualified with the parameter
    type used for tests.

    Parameters
    ----------
    job_summary_type
        Type of XML job summary specialized for the XML model used for job
        parameters. For example, ``JobSummary[SomeXmlParameters]``.
    seen
        XML returned by the application under test.
    expected
        Expected XML.
    """
    seen_model = job_summary_type.from_xml(seen)
    expected_model = job_summary_type.from_xml(expected)
    assert seen_model.model_dump() == expected_model.model_dump()



"""Helper code for FastAPI (other than dependencies and middleware)."""

from ._errors import ClientRequestError, client_request_error_handler

__all__ = [
    "ClientRequestError",
    "client_request_error_handler",
]



"""An asyncio multiple reader, multiple writer queue."""

import asyncio
from collections.abc import AsyncGenerator
from datetime import UTC, datetime, timedelta
from types import EllipsisType

__all__ = [
    "AsyncMultiQueue",
    "AsyncMultiQueueError",
]


class AsyncMultiQueueError(Exception):
    """Invalid sequence of calls when writing to `AsyncMultiQueue`."""


class AsyncMultiQueue[T]:
    """An asyncio multiple reader, multiple writer queue.

    Provides a generic queue for asyncio that supports multiple readers (via
    async iterator) and multiple writers. Readers can start reading at any
    time and will begin reading from the start of the queue. There is no
    maximum size of the queue; new items will be added subject only to the
    limits of available memory.

    This data structure is not thread-safe. It uses only asyncio locking, not
    thread-safe locking.

    The ellipsis object (``...``) is used as a placeholder to indicate the end
    of the queue, so cannot be pushed onto the queue.
    """

    def __init__(self) -> None:
        self._contents: list[T | EllipsisType] = []
        self._triggers: list[asyncio.Event] = []

    def __aiter__(self) -> AsyncGenerator[T]:
        """Return an async iterator over the queue."""
        return self.aiter_from(0)

    @property
    def finished(self) -> bool:
        """Whether `close` has been called on the queue.

        If this property is `True`, the contents of the queue are finalized
        and no new items will be added unless the queue is cleared with
        `clear`.
        """
        return bool(self._contents and self._contents[-1] is Ellipsis)

    def aiter_from(
        self, start: int, timeout: timedelta | None = None
    ) -> AsyncGenerator[T]:
        """Return an async iterator over the queue.

        Each call to this function returns a separate iterator over the same
        underlying contents, and each iterator will be triggered separately.

        Parameters
        ----------
        start
            Starting position in the queue. This can be larger than the
            current queue size, in which case no items are returned until the
            queue passes the given starting position.
        timeout
            If given, total length of time for the iterator. This isn't the
            timeout waiting for the next item; this is the total execution
            time of the iterator.

        Returns
        -------
        AsyncGenerator
            An async generator over the contents of the queue.

        Raises
        ------
        TimeoutError
            Raised when the timeout is reached.
        """
        end_time = datetime.now(tz=UTC) + timeout if timeout else None

        # Grab a reference to the current contents so that the iterator
        # detaches from the contents on clear.
        contents = self._contents

        # Add a trigger for this caller.
        trigger = asyncio.Event()
        self._triggers.append(trigger)

        # Construct the iterator, which waits for the trigger and returns any
        # new events until it sees the placeholder for the end of the queue
        # (the ellipsis object).
        async def iterator() -> AsyncGenerator[T]:
            position = start
            try:
                while True:
                    trigger.clear()
                    end = len(contents)
                    if position < end:
                        for item in contents[position:end]:
                            if item is Ellipsis:
                                return
                            yield item
                        position = end
                    elif contents and contents[-1] is Ellipsis:
                        return
                    if end_time:
                        now = datetime.now(tz=UTC)
                        timeout_left = (end_time - now).total_seconds()
                        async with asyncio.timeout(timeout_left):
                            await trigger.wait()
                    else:
                        await trigger.wait()
            finally:
                self._triggers = [t for t in self._triggers if t != trigger]

        return iterator()

    def clear(self) -> None:
        """Empty the contents of the queue.

        Any existing readers will still see all items pushed to the queue
        before the clear, but will become detached from the queue and will not
        see any new events added after the clear.
        """
        finished = self.finished
        contents = self._contents
        triggers = self._triggers
        self._contents = []
        self._triggers = []
        if not finished:
            contents.append(Ellipsis)
            for trigger in triggers:
                trigger.set()

    def close(self) -> None:
        """Mark the end of the queue data.

        Similar to `clear` in that any existing readers of the queue will see
        the end of the iterator, but the data will not be deleted and new
        readers can still read all of the data in the queue.

        After `close` is called, `clear` must be called before any subsequent
        `put`.
        """
        self._contents.append(Ellipsis)
        for trigger in self._triggers:
            trigger.set()

    def put(self, item: T) -> None:
        """Add an item to the queue.

        Parameters
        ----------
        item
           Item to add.

        Raises
        ------
        AsyncMultiQueueError
            Raised if `put` was called after `close` without an intervening
            call to `clear`.
        """
        if self.finished:
            msg = "end was already called, must call clear before put"
            raise AsyncMultiQueueError(msg)
        self._contents.append(item)
        for trigger in self._triggers:
            trigger.set()

    def qsize(self) -> int:
        """Return the number of items currently in the queue."""
        count = len(self._contents)
        if self.finished:
            count -= 1
        return count



"""Utiility functions for managing an external Uvicorn test process.

Normally, ASGI apps are tested via the built-in support in HTTPX for running
an ASGI app directly. However, sometimes the app has to be spawned in a
separate process so that it can be accessed over HTTP, such as when testing it
with Selenium or when testing Uvicorn integration. This module provides
utility functions to aid with that test setup.
"""

from __future__ import annotations

import errno
import os
import socket
import subprocess
import time
from dataclasses import dataclass
from pathlib import Path

import structlog
from structlog.stdlib import BoundLogger

__all__ = [
    "ServerNotListeningError",
    "UvicornProcess",
    "spawn_uvicorn",
]


class ServerNotListeningError(Exception):
    """Timeout waiting for the server to start listening."""


@dataclass
class UvicornProcess:
    """Properties of the running Uvicorn service."""

    process: subprocess.Popen
    """Running Uvicorn process."""

    url: str
    """URL (on localhost) on which the process is listening."""


def _wait_for_server(port: int, timeout: float = 5.0) -> None:
    """Wait until a server accepts connections on the specified port.

    Parameters
    ----------
    port
        Port on localhost on which the server will listen.
    timeout
        How long to wait in seconds.

    Raises
    ------
    ServerNotListeningError
        The server did not start within the specified timeout.
    """
    deadline = time.time() + timeout
    while True:
        socket_timeout = deadline - time.time()
        if socket_timeout < 0.0:
            msg = f"Server did not start on port {port} in {timeout}s"
            raise ServerNotListeningError(msg)
        try:
            sock = socket.socket()
            sock.settimeout(socket_timeout)
            sock.connect(("localhost", port))
        except TimeoutError:
            pass
        except OSError as e:
            if e.errno not in (errno.ETIMEDOUT, errno.ECONNREFUSED):
                raise
        else:
            sock.close()
            return
        time.sleep(0.1)


def spawn_uvicorn(
    *,
    working_directory: str | Path,
    app: str | None = None,
    factory: str | None = None,
    capture: bool = False,
    timeout: float = 5.0,
    env: dict[str, str] | None = None,
    logger: BoundLogger | None = None,
) -> UvicornProcess:
    """Spawn an ASGI app as a separate Uvicorn process.

    The current working directory will always be added to the Python path, so
    ``app`` and ``factory`` can point to modules found relative to the current
    working directory, or relative to the ``working_directory`` parameter.

    Parameters
    ----------
    working_directory
        Directory in which Uvicorn should run. Normally this should come from
        the ``tmp_path`` fixture.
    app
        Module and variable name of the app. Either this or ``factory`` must
        be given.
    factory
        Module name and callable that will create the app object. Either this
        or ``app`` must be given.
    capture
        Whether to capture standard output and standard error of Uvicorn. If
        set to true, both will be automatically decoded as UTF-8 text and will
        be available in the ``process`` attribute of the returned object.
    timeout
        How long to wait in seconds.
    env
        Extra environment variable settings.
    logger
        Logger to use for logging, if given. If not, a new structlog logger
        will be created.

    Returns
    -------
    UvicornProcess
        Properties of the running Uvicorn service.

    Raises
    ------
    RuntimeError
        Raised if no logger was provided and no logger could be obtained from
        structlog.
    ServerNotListeningError
        Raised if the server did not start within the timeout. Generally this
        means it had some fatal error during startup.
    ValueError
        Raised if either both or neither of app and factory were given.
    """
    if app and factory:
        raise ValueError("Only one of app or factory may be given")
    if not app and not factory:
        raise ValueError("Neither of app nor factory was given")
    if not logger:
        logger = structlog.get_logger("safir.testing.uvicorn")
        if not logger:
            raise RuntimeError("Cannot get logger")
    env = {**os.environ, **env} if env else {**os.environ}

    # Get a random port for the app to listen on.
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]

    # Spawn the app.
    cmd = ["uvicorn", "--fd", "0"]
    if app:
        cmd.append(app)
    elif factory:
        cmd.extend(("--factory", factory))
    if "PYTHONPATH" in env:
        env["PYTHONPATH"] += f":{Path.cwd()}"
    else:
        env["PYTHONPATH"] = str(Path.cwd())
    logger.info("Starting server", command=cmd)
    if capture:
        process = subprocess.Popen(
            cmd,
            cwd=str(working_directory),
            stdin=sock.fileno(),
            env=env,
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    else:
        process = subprocess.Popen(
            cmd,
            cwd=str(working_directory),
            stdin=sock.fileno(),
            env=env,
            text=True,
        )
    sock.close()

    # Wait for it to start listening.
    logger.info("Waiting for server to start")
    _wait_for_server(port, timeout)

    # Server successfully started. Return its properties.
    return UvicornProcess(process=process, url=f"http://127.0.0.1:{port}")



"""Support for testing Kafka helpers."""

import logging
from collections.abc import AsyncGenerator, Generator
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from pathlib import Path

from aiokafka import AIOKafkaConsumer
from aiokafka.admin import AIOKafkaAdminClient
from docker.errors import NotFound
from faststream.kafka import KafkaBroker
from pydantic import AnyUrl
from schema_registry.client import AsyncSchemaRegistryClient
from testcontainers.core.network import Network

from safir.kafka import (
    KafkaConnectionSettings,
    SchemaManagerSettings,
    SecurityProtocol,
)
from safir.kafka._manager import PydanticSchemaManager
from safir.testing.containers import (
    FullKafkaContainer,
    SchemaRegistryContainer,
)

__all__ = [
    "KafkaClients",
    "KafkaStack",
    "make_kafka_clients",
    "make_kafka_stack",
]


@dataclass
class KafkaStack:
    """Objects and external services in a full app metrics stack."""

    kafka_container: FullKafkaContainer
    """A Kafka docker TestContainer."""

    kafka_cert_path: Path
    """The local path containing the Kafka certs."""

    schema_registry_container: SchemaRegistryContainer
    """A schema registry TestContainer."""

    kafka_connection_settings: KafkaConnectionSettings
    """Connection settings for the Kafka container."""

    schema_manager_settings: SchemaManagerSettings
    """Settings for a schema manager pointed at the stack."""


@dataclass
class KafkaClients:
    """Clients for pieces of the Kafka stack."""

    kafka_consumer: AIOKafkaConsumer
    """An AIOKafkaConsumer pointed at the stack."""

    kafka_broker: KafkaBroker
    """A faststream KafkaBroker pointed at the stack."""

    kafka_admin_client: AIOKafkaAdminClient
    """An AIOKafkaAdminClient pointed at the stack."""

    schema_registry_client: AsyncSchemaRegistryClient
    """A schema registry client pointed at the stack."""

    schema_manager: PydanticSchemaManager
    """A PydanticSchemaManager pointed at the stack."""


@contextmanager
def make_kafka_stack(
    kafka_cert_path: Path, *, constant_host_ports: bool = False
) -> Generator[KafkaStack]:
    """Yield a full kafka stack with clients and underlying infrastructure.

    Parameters
    ----------
    kafka_cert_path
        The local path to Kafka cert files. Probably comes from a
        tmp_path_factory fixture.
    constant_host_ports
        Whether or not the Kafka container should use explicitly numbered host
        ports. This is needed for tests that stop and restart the Kafka
        container so that it comes back up with the same mapped host ports.
    """
    kafka_container = None
    schema_registry_container = None
    network = None
    try:
        network = Network().create()
        kafka_container = FullKafkaContainer(
            constant_host_ports=constant_host_ports
        )
        kafka_container.with_network(network)
        kafka_container.with_network_aliases("kafka")

        kafka_container.start()

        for filename in ("ca.crt", "client.crt", "client.key"):
            contents = kafka_container.get_secret_file_contents(filename)
            (kafka_cert_path / filename).write_text(contents)

        kafka_connection_settings = KafkaConnectionSettings(
            bootstrap_servers=kafka_container.get_bootstrap_server(),
            security_protocol=SecurityProtocol.PLAINTEXT,
        )
        schema_registry_container = SchemaRegistryContainer(network=network)
        schema_registry_container.with_network(network)
        schema_registry_container.with_network_aliases("schemaregistry")

        schema_registry_container.start()

        schema_manager_settings = SchemaManagerSettings(
            registry_url=AnyUrl(schema_registry_container.get_url())
        )

        yield KafkaStack(
            kafka_container=kafka_container,
            kafka_cert_path=kafka_cert_path,
            schema_registry_container=schema_registry_container,
            kafka_connection_settings=kafka_connection_settings,
            schema_manager_settings=schema_manager_settings,
        )

    finally:
        if kafka_container:
            try:
                kafka_container.stop()
            except NotFound:
                logging.getLogger().info(
                    "Trying to stop Kafka container, but it doesn't exist."
                    " This is fine if the container was stopped in the test"
                    " itself."
                )

        if schema_registry_container:
            try:
                schema_registry_container.stop()
            except NotFound:
                logging.getLogger().info(
                    "Trying to stop Schema Registry container, but it doesn't"
                    " exist. This is fine if the container was stopped in the"
                    " test itself."
                )
        if network:
            network.remove()


@asynccontextmanager
async def make_kafka_clients(
    stack: KafkaStack,
) -> AsyncGenerator[KafkaClients]:
    """Yield a bucket of useful Kafka and schema manager clients.

    Parameters
    ----------
    stack
        The containers to point the clients at.
    """
    kafka_consumer = None
    kafka_broker = None
    kafka_admin_client = None

    try:
        schema_registry_client = AsyncSchemaRegistryClient(
            **stack.schema_manager_settings.to_registry_params()
        )

        kafka_consumer = AIOKafkaConsumer(
            **stack.kafka_connection_settings.to_aiokafka_params(),
            client_id="pytest-consumer",
        )
        await kafka_consumer.start()

        kafka_broker = KafkaBroker(
            **stack.kafka_connection_settings.to_faststream_params(),
            client_id="pytest-broker",
        )
        await kafka_broker.start()

        kafka_admin_client = AIOKafkaAdminClient(
            **stack.kafka_connection_settings.to_aiokafka_params(),
            client_id="pytest-admin",
        )
        await kafka_admin_client.start()

        schema_manager = stack.schema_manager_settings.make_manager()

        yield KafkaClients(
            schema_registry_client=schema_registry_client,
            kafka_consumer=kafka_consumer,
            kafka_broker=kafka_broker,
            kafka_admin_client=kafka_admin_client,
            schema_manager=schema_manager,
        )
    finally:
        if kafka_consumer:
            await kafka_consumer.stop()
        if kafka_broker:
            await kafka_broker.stop()
        if kafka_admin_client:
            await kafka_admin_client.close()



"""Test the logger FastAPI dependency."""

from __future__ import annotations

import json
from typing import Annotated
from unittest.mock import ANY

import pytest
from fastapi import Depends, FastAPI
from httpx import ASGITransport, AsyncClient
from structlog.stdlib import BoundLogger

from safir.dependencies.logger import logger_dependency
from safir.logging import configure_logging
from safir.middleware.x_forwarded import XForwardedMiddleware


@pytest.mark.asyncio
async def test_logger(caplog: pytest.LogCaptureFixture) -> None:
    configure_logging(name="myapp", profile="production", log_level="info")

    app = FastAPI()

    @app.get("/")
    async def handler(
        logger: Annotated[BoundLogger, Depends(logger_dependency)],
    ) -> dict[str, str]:
        logger.info("something", param="value")
        return {}

    caplog.clear()
    transport = ASGITransport(app=app)
    base_url = "http://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get(
            "/", headers={"User-Agent": "some-user-agent/1.0"}
        )
    assert r.status_code == 200

    assert len(caplog.record_tuples) == 1
    assert json.loads(caplog.record_tuples[0][2]) == {
        "event": "something",
        "httpRequest": {
            "requestMethod": "GET",
            "requestUrl": "http://example.com/",
            "remoteIp": "127.0.0.1",
            "userAgent": "some-user-agent/1.0",
        },
        "logger": "myapp",
        "param": "value",
        "request_id": ANY,
        "severity": "info",
    }


@pytest.mark.asyncio
async def test_logger_xforwarded(caplog: pytest.LogCaptureFixture) -> None:
    configure_logging(name="myapp", profile="production", log_level="info")

    app = FastAPI()
    app.add_middleware(XForwardedMiddleware)

    @app.get("/")
    async def handler(
        logger: Annotated[BoundLogger, Depends(logger_dependency)],
    ) -> dict[str, str]:
        logger.info("something", param="value")
        return {}

    caplog.clear()
    transport = ASGITransport(app=app)
    base_url = "https://example.com"
    async with AsyncClient(transport=transport, base_url=base_url) as client:
        r = await client.get(
            "/",
            headers={
                "Host": "foo.example.com",
                "User-Agent": "",
                "X-Forwarded-For": "10.10.10.10",
                "X-Forwarded-Proto": "https",
                "X-Forwarded-Host": "foo.example.com",
            },
        )
    assert r.status_code == 200

    assert len(caplog.record_tuples) == 1
    assert json.loads(caplog.record_tuples[0][2]) == {
        "event": "something",
        "httpRequest": {
            "requestMethod": "GET",
            "requestUrl": "https://foo.example.com/",
            "remoteIp": "10.10.10.10",
        },
        "logger": "myapp",
        "param": "value",
        "request_id": ANY,
        "severity": "info",
    }



"""Safir-provided test containers."""

from ._kafka import FullKafkaContainer
from ._schema_registry import SchemaRegistryContainer

__all__ = [
    "FullKafkaContainer",
    "SchemaRegistryContainer",
]
